import struct
from abc import ABCMeta, abstractmethod


def bytes2binstr(b, n=None):
    s = ''.join(f'{x:08b}' for x in b)
    return s if n is None else s[:n + n // 8 + (0 if n % 8 else -1)]


def getbytes(bits):
    done = False
    while not done:
        byte = 0
        for _ in range(0, 8):
            try:
                bit = next(bits)
            except StopIteration:
                bit = 0
                done = True
            byte = (byte << 1) | bit
        yield byte


class EEPROMField(object):
    '''
    EEPROM Field
    '''

    def __init__(self, name, address, word_len=1):
        self.name = name
        self.address = address
        self.word_len = word_len

    @abstractmethod
    def parse(self, payload):
        '''
        Parse payload
        '''
        # return: value, parsed, error
        return payload, False, None


class ProductConfigurationField(EEPROMField):
    def __init__(self, name, address, word_len=1):
        super(ProductConfigurationField, self).__init__(
            name, address, word_len=1)

    def parse(self, payload):
        # return: value, parsed, error
        bytes_value = struct.pack('BB', *payload)
        bit_value = bytes2binstr(bytes_value)

        parsed_value = {
            'mags':int(bit_value[-1]),
            'gps':int(bit_value[-2]),
            'algorithm':int(bit_value[-3]),
            'ext_aiding':int(bit_value[-4]),
            'architechture':int(bit_value[-8:-4], 2)
        }

        return parsed_value, True, None


EEPROM_ADDRESS_DEFINES = [
    {
        "address": 0x71C,
        "instance": ProductConfigurationField('Product Configuration', 0x71C)
    }
]


class EEPROMFieldDefines:
    '''
    A list of fields could be parsed
    '''

    def __init__(self):
        self._list = dict()
        self._default_field = EEPROMField('default', -1)

    def load(self):
        self._list.clear()
        for item in EEPROM_ADDRESS_DEFINES:
            self._list[item['address']] = item['instance']

    def find(self, address: int) -> EEPROMField:
        exist_field = self._list.get(address)

        if not exist_field:
            exist_field = self._default_field

        return exist_field


EEPROM_FIELD_DEFINES_SINGLETON = EEPROMFieldDefines()



from abc import ABCMeta, abstractmethod
import os
import sys
import threading
import collections
import time
import struct
import traceback
from pathlib import Path
from azure.storage.blob import BlockBlobService
from .event_base import EventBase
from ...framework.utils import (helper, resource)
from ...framework.file_storage import FileLoger
from ...framework.configuration import get_config
from ..message_center import DeviceMessageCenter
from ..parser_manager import ParserManager

if sys.version_info[0] > 2:
    from queue import Queue
else:
    from Queue import Queue


class OpenDeviceBase(EventBase):
    '''
    Base class of open device(openimu, openrtk)
    '''
    __metaclass__ = ABCMeta

    def __init__(self, communicator):
        super(OpenDeviceBase, self).__init__()
        self.type = 'None'
        self.threads = []  # thread of receiver and paser
        self.exception_thread = False  # flag of exit threads
        self.exception_lock = threading.Lock()  # lock of exception_thread
        self.exit_thread = False
        self.data_queue = Queue()  # data container
        self.data_lock = threading.Lock()
        self.clients = []
        self.input_result = None
        self.bootloader_result = None
        self.is_streaming = False
        self.has_running_checker = False
        self.has_backup_checker = False
        self.connected = False
        self.is_upgrading = False
        self.complete_upgrade = False
        self.communicator = communicator
        self.bootloader_baudrate = 57600
        self.properties = None
        self.firmware_content = []
        self.fs_len = 0
        self.addr = 0
        self.max_data_len = 240
        self.block_blob_service = None
        self._logger = None
        self.is_logging = False
        self.enable_data_log = True
        self.cli_options = None
        self._message_center = None

    @abstractmethod
    def load_properties(self):
        '''
        load configuration
        '''

    @abstractmethod
    def on_receive_output_packet(self, packet_type, data):
        '''
        Listener for receiving output packet
        '''

    @abstractmethod
    def after_setup(self):
        pass

    def internal_input_command(self, command, read_length=500):
        '''
        Internal input command
        '''
        command_line = helper.build_input_packet(command)
        self.communicator.write(command_line)
        time.sleep(0.1)

        data_buffer = self.read_untils_have_data(command, read_length, 20)
        parsed = bytearray(data_buffer) if data_buffer and len(
            data_buffer) > 0 else None

        format_string = None
        if parsed is not None:
            try:
                if sys.version_info < (3, 0):
                    format_string = str(struct.pack(
                        '{0}B'.format(len(parsed)), *parsed))
                else:
                    format_string = str(struct.pack(
                        '{0}B'.format(len(parsed)), *parsed), 'utf-8')
            except UnicodeDecodeError:
                return ''

        if format_string is not None:
            return format_string
        return ''

    def _parse_buffer(self, data_buffer):
        response = {
            'parsed': False,
            'parsed_end_index': 0,
            'result': []
        }
        data_queue = Queue()
        data_queue.queue.extend(data_buffer)

        command_start = [0x55, 0x55]
        parsed_data = []
        is_header_found = False
        packet_type = ''
        data_buffer_len = len(data_buffer)

        while not data_queue.empty():
            if is_header_found:
                # if matched packet, is_header_found = False, parsed_data = []
                if not data_queue.empty():
                    packet_type_start = data_queue.get()
                else:
                    break

                if not data_queue.empty():
                    packet_type_end = data_queue.get()
                else:
                    break

                if not data_queue.empty():
                    packet_len = data_queue.get()
                    packet_type = ''.join(
                        ["%c" % x for x in [packet_type_start, packet_type_end]])
                    packet_data = []

                    if data_queue.qsize() >= packet_len:
                        # take packet
                        for _ in range(packet_len):
                            packet_data.append(data_queue.get())
                    else:
                        break
                    # update response
                    response['parsed'] = True
                    response['result'].append({
                        'type': packet_type,
                        'data': packet_data
                    })
                    response['parsed_end_index'] += data_buffer_len - \
                        data_queue.qsize()
                    data_buffer_len = data_queue.qsize()
                    parsed_data = []
                    is_header_found = False
                else:
                    break
            else:
                byte_item = data_queue.get()
                parsed_data.append(byte_item)

                if len(parsed_data) > 2:
                    parsed_data = parsed_data[-2:]

                if parsed_data == command_start:
                    # find message start
                    is_header_found = True
                    parsed_data = []

        return response

    # may lost data
    def read_untils_have_data(self, packet_type, read_length=200, retry_times=20):
        '''
        Get data from limit times of read
        '''
        result = None
        trys = 0
        data_buffer = []
        self.data_queue.queue.clear()

        while trys < retry_times:
            data_buffer_per_time = bytearray(
                self.communicator.read(read_length))
            data_buffer.extend(data_buffer_per_time)

            response = self._parse_buffer(data_buffer)
            if response['parsed']:
                matched_packet = next(
                    (packet['data'] for packet in response['result'] if packet['type'] == packet_type), None)
                if matched_packet is not None:
                    result = matched_packet
                    break
                else:
                    # clear buffer to parsed index
                    data_buffer = data_buffer[response['parsed_end_index']:]
            trys += 1

        return result

    def _setup_message_center(self):
        if not self._message_center:
            self._message_center = DeviceMessageCenter(self.communicator)

        if not self._message_center.is_ready():
            uart_parser = ParserManager.build(self.type, self.properties)
            self._message_center.set_parser(uart_parser)
            self._message_center.on(
                'continuous_message',
                self.on_receive_continuous_messsage
            )
            self._message_center.on(
                'error',
                self.on_recevie_message_center_error
            )
            self._message_center.on(
                'read_block',
                self.on_read_raw
            )
            self._message_center.setup()
        else:
            self._message_center.get_parser().set_configuration(self.properties)
            self._message_center.setup()

    def setup(self, options):
        ''' start 2 threads, receiver, parser
        '''
        self.load_properties()
        self._logger = FileLoger(self.properties)
        self.cli_options = options

        with_data_log = options and options.with_data_log

        self._setup_message_center()

        if with_data_log and not self.is_logging and self.enable_data_log:
            log_result = self._logger.start_user_log('data')
            if log_result == 1 or log_result == 2:
                raise Exception('Cannot start data logger')
            self.is_logging = True

        self.after_setup()

    def on_recevie_message_center_error(self, error_type, message):
        '''
        event handler after got message center error
        '''
        self.emit('exception', error_type, message)

    def on_receive_continuous_messsage(self, packet_type, data):
        '''
        event handler after got continuous message
        '''
        if isinstance(data, list):
            for item in data:
                self._logger.append(packet_type, item)
        else:
            self._logger.append(packet_type, data)

        self.on_receive_output_packet(packet_type, data)

    @abstractmethod
    def on_read_raw(self, data):
        pass

    def get_command_lines(self):
        '''
        Get command line defines
        '''
        if self.properties.__contains__('CLICommands'):
            return self.properties['CLICommands']
        return []

    def add_output_packet(self, method, packet_type, data):
        '''
        Add output packet
        '''
        for client in self.clients:
            client.on_receive_output_packet(method, packet_type, data)

    def append_client(self, client):
        '''
        Append client connection, cache it
        '''
        self.clients.append(client)

    def remove_client(self, client):
        '''
        Remove specified client
        '''
        self._reset_client()
        self.clients.remove(client)

    def _reset_client(self):
        self.input_result = None
        self.bootloader_result = None
        self.is_streaming = False
        self.is_upgrading = False

        self._message_center.resume()

        if self._logger is not None:
            self._logger.stop_user_log()
            self.is_logging = False

    def reset(self):
        '''
        Reset
        '''
        self._reset_client()
        # self.threads.clear()
        helper.clear_elements(self.threads)
        self.listeners.clear()
        # self.clients.clear()
        self.exception_thread = False
        # self.data_queue.queue.clear()

    def close(self):
        '''
        Close and disconnect
        '''
        helper.clear_elements(self.clients)
        self.reset()
        self._message_center.stop()
        self._message_center = None
        self.exit_thread = True

    def restart(self):
        '''
        Restart device
        '''
        # output firmware upgrade finished
        time.sleep(1)
        command_line = helper.build_bootloader_input_packet('JA')
        self.communicator.write(command_line)
        print('Restarting app ...')
        time.sleep(5)

        self.emit('complete_upgrade')

    def thread_do_upgrade_framework(self, file):
        '''
        Do upgrade firmware
        '''
        try:
            # step.1 download firmware
            if not self.download_firmware(file):
                self.on_upgarde_failed('cannot find firmware file')
                return
            # step.2 jump to bootloader
            if not self.switch_to_bootloader():
                self.on_upgarde_failed('Bootloader Start Failed')
                return
            # step.3 write to block
            print('Firmware upgrading...')
            self.write_firmware()
            # step.4 restart app
            self.restart()
        except Exception:  # pylint:disable=broad-except
            self.on_upgarde_failed('Upgrade Failed')
            traceback.print_exc()

    def _do_download_firmware(self, file):
        upgarde_root = os.path.join(resource.get_executor_path(), 'upgrade')
        firmware_content = None

        if not os.path.exists(upgarde_root):
            os.makedirs(upgarde_root)

        firmware_file_path = os.path.join(upgarde_root, file)
        firmware_file = Path(firmware_file_path)

        config = get_config()

        if firmware_file.is_file():
            firmware_content = open(firmware_file_path, 'rb').read()
        else:
            self.block_blob_service = BlockBlobService(
                account_name=config.AZURE_STORAGE_ACCOUNT, protocol='https')
            self.block_blob_service.get_blob_to_path(
                config.AZURE_STORAGE_APPS_CONTAINER, file, firmware_file_path)
            firmware_content = open(firmware_file_path, 'rb').read()

        return firmware_content

    def download_firmware(self, file):
        '''
        Downlaod firmware from Azure storage
        '''
        can_download = False
        try:
            self.firmware_content = self._do_download_firmware(file)
            self.addr = 0
            self.fs_len = len(self.firmware_content)
            can_download = True
        except Exception as e:
            can_download = False

        return can_download

    def switch_to_bootloader(self):
        '''
        Switch to bootloader
        '''
        try:
            # TODO: should send set quiet command before go to bootloader mode
            command_line = helper.build_bootloader_input_packet('JI')
            self.communicator.reset_buffer()  # clear input and output buffer
            self.communicator.write(command_line, True)
            time.sleep(3)
            # It is used to skip streaming data with size 1000 per read
            self.read_untils_have_data('JI', 1000, 50)
            self.communicator.serial_port.baudrate = self.bootloader_baudrate
            return True
        except Exception as ex:  # pylint:disable=broad-except
            print('bootloader exception', ex)
            return False

    def write_firmware(self):
        '''Upgrades firmware of connected device to file provided in argument
        '''
        while self.addr < self.fs_len:
            packet_data_len = self.max_data_len if (
                self.fs_len - self.addr) > self.max_data_len else (self.fs_len - self.addr)
            data = self.firmware_content[self.addr: (
                self.addr + packet_data_len)]
            self.write_block(packet_data_len, self.addr, data)
            self.addr += packet_data_len
            self.add_output_packet('stream', 'upgrade_progress', {
                'addr': self.addr,
                'fs_len': self.fs_len
            })
            # output firmware upgrading

    def write_block(self, data_len, addr, data):
        '''
        Send block to bootloader
        '''
        # print(data_len, addr, time.time())
        command_line = helper.build_bootloader_input_packet(
            'WA', data_len, addr, data)
        try:
            self.communicator.write(command_line, True)
        except Exception:  # pylint: disable=broad-except
            self.exception_lock.acquire()
            self.exception_thread = True
            self.exception_lock.release()
            return

        if addr == 0:
            time.sleep(8)

        response = self.read_untils_have_data('WA', 50, 50)
        # wait WA end if cannot read response in defined retry times
        if response is None:
            time.sleep(0.1)

    def upgrade_completed(self, options):
        '''
        Actions after upgrade complete
        '''
        self.input_result = None
        self.bootloader_result = None
        # self.data_queue.queue.clear()
        self.is_upgrading = False

        self.load_properties()
        self._logger = FileLoger(self.properties)
        self.cli_options = options

        self._message_center.get_parser().set_configuration(self.properties)
        self._message_center.resume()

        if options and not options.with_data_log and self.enable_data_log:
            log_result = self._logger.start_user_log('data')
            if log_result == 1 or log_result == 2:
                raise Exception('Cannot start data logger')
            self.is_logging = True

    def start_data_log(self):
        '''
        Start to log
        '''
        if self.is_logging:
            return False

        if self._logger is None:
            self._logger = FileLoger(self.properties)

        log_result = self._logger.start_user_log('data')
        if log_result == 1 or log_result == 2:
            raise Exception('Cannot start data logger')
        self.is_logging = True

    def stop_data_log(self):
        '''
        Stop logging
        '''
        if self.is_logging and not self._logger:
            self._logger.stop_user_log()
            self.is_logging = False

    def on_upgarde_failed(self, message):
        '''
        Linstener for upgrade failure
        '''
        self.is_upgrading = False
        self._message_center.resume()
        self.add_output_packet(
            'stream', 'upgrade_complete', {'success': False, 'message': message})



class EventBase(object):
    '''
    Event Object Base
    '''

    def __init__(self):
        self.listeners = {}

    def on(self, event_type, handler):
        '''
        Listen event
        '''
        if not self.listeners.__contains__(event_type):
            self.listeners[event_type] = []

        self.listeners[event_type].append(handler)

    def emit(self, event_type, *args, **kwargs):
        '''
        Trigger event
        '''
        if not self.listeners.keys().__contains__(event_type):
            return

        handlers = self.listeners[event_type]
        if handlers is not None and len(handlers) > 0:
            for handler in handlers:
                handler(*args, **kwargs)



import os
import sys
import argparse

try:
    from aceinna.bootstrap.cli import CommandLine
    from aceinna.framework.constants import BAUDRATE_LIST
except:  # pylint: disable=bare-except
    print('load package from local')
    sys.path.append('./src')
    from aceinna.bootstrap.cli import CommandLine
    from aceinna.framework.constants import BAUDRATE_LIST


def receive_args():
    """parse input arguments
    """
    parser = argparse.ArgumentParser(
        description='Aceinna python driver input args command:')
    # parser.add_argument("-host", type=str, help="host type", default='web')
    # for host as web
    parser.add_argument("-p", "--port", type=int,
                        help="Webserver port")
    parser.add_argument("--device-type", type=str,
                        help="Open Device Type")
    parser.add_argument("-b", "--baudrate", type=int,
                        help="Baudrate for uart", choices=BAUDRATE_LIST)
    parser.add_argument("-c", "--com-port", type=str,
                        help="COM Port")
    parser.add_argument("--console-log", dest='console_log', action='store_true',
                        help="Output log on console", default=False)
    parser.add_argument("--debug", dest='debug', action='store_true',
                        help="Log debug information", default=False)
    parser.add_argument("--with-data-log", dest='with_data_log', action='store_true',
                        help="Contains internal data log (OpenIMU only)", default=False)
    parser.add_argument("--with-raw-log", dest='with_raw_log', action='store_true',
                        help="Contains raw data log (OpenRTK only)", default=False)
    return parser.parse_args()


def main():
    '''start'''
    input_args = receive_args()
    command_line = CommandLine(
        device_type=input_args.device_type,
        com_port=input_args.com_port,
        port=input_args.port,
        baudrate=input_args.baudrate,
        console_log=input_args.console_log,
        debug=input_args.debug,
        with_data_log=input_args.with_data_log,
        with_raw_log=input_args.with_raw_log
    )
    command_line.listen()


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:  # response for KeyboardInterrupt such as Ctrl+C
        print('User stop this program by KeyboardInterrupt! File:[{0}], Line:[{1}]'.format(
            __file__, sys._getframe().f_lineno))
        sys.exit()
    except:  # pylint: disable=bare-except
        os._exit(1)



import struct
from typing import List
from .eeprom_field import EEPROMField
from .configuration_field import ConfigurationField
from ...framework.utils import helper

COMMAND_START = [0x55, 0x55]


def build_read_field_packets(field: ConfigurationField, write_to_eeprom=False) -> bytearray:
    message_type = 'GF' if not write_to_eeprom else 'RF'

    fields_bytes = []
    field_id_bytes = struct.pack('>h', field.field_id)
    fields_bytes.extend(field_id_bytes)

    return build_packet(message_type, fields_bytes)


def build_read_fields_packets(fields: List[ConfigurationField], write_to_eeprom=False) -> bytearray:
    message_type = 'GF' if not write_to_eeprom else 'RF'

    fields_bytes = []
    for field in fields:
        field_id_bytes = struct.pack('>h', field.field_id)
        fields_bytes.extend(field_id_bytes)

    return build_packet(message_type, fields_bytes)


def build_write_filed_cli(field: ConfigurationField, value, write_to_eeprom=False) -> bytearray:
    message_type = 'SF' if not write_to_eeprom else 'WF'

    fields_bytes = []

    field_id_bytes = struct.pack('>h', field.field_id)
    field_value_bytes = field.encode(value)

    fields_bytes.extend(field_id_bytes)
    fields_bytes.extend(field_value_bytes)

    return build_packet(message_type, fields_bytes)


def build_write_fileds_cli(fields: List[ConfigurationField], values, write_to_eeprom=False) -> bytearray:
    message_type = 'SF' if not write_to_eeprom else 'WF'

    fields_bytes = []
    for index, field in enumerate(fields):
        field_id_bytes = struct.pack('>h', field.field_id)
        field_value_bytes = field.encode(values[index])

        fields_bytes.extend(field_id_bytes)
        fields_bytes.extend(field_value_bytes)

    return build_packet(message_type, fields_bytes)


def build_read_eeprom_cli(field: EEPROMField) -> bytearray:
    fields_bytes = []

    start_address_bytes = struct.pack('>h', field.address)

    fields_bytes.extend(start_address_bytes)
    fields_bytes.extend([field.word_len])

    return build_packet('RE', fields_bytes)


def build_write_eeproms_cli():
    pass


def build_packet(message_type, message_bytes=[]):
    '''
    build dmu command packet
    inspired from https://github.com/rishitborad/IMU383_Verification_Scripts/blob/master/IMU383_Uart.py
    '''
    packet = []
    packet.extend(bytearray(message_type, 'utf-8'))

    if message_type == "WF" or message_type == "SF":
        msg_len = 1 + len(message_bytes)
        no_of_fields = int(len(message_bytes)/4)
        packet.append(msg_len)
        packet.append(no_of_fields)
        final_packet = packet + message_bytes
        # print packet
    elif message_type == "RF" or message_type == "GF":
        msg_len = 1 + len(message_bytes)
        no_of_fields = int(len(message_bytes)/2)
        packet.append(msg_len)
        packet.append(no_of_fields)
        final_packet = packet + message_bytes
    else:
        msg_len = len(message_bytes)
        packet.append(msg_len)
        final_packet = packet + message_bytes
    # print(message_type, final_packet)
    return COMMAND_START + final_packet + helper.calc_crc(final_packet)


def build_continous_packet_types(architecture, algorithm, mags):
    '''
    Output Packets      architecture	        algorithm	mags
    S0	                any except 4 and 5	    any	        1
    S2	                any except 4 and 5	    1	        any
    S3	                4 or 5	                any	        any
    A0,A1,	            any except 4 and 5 	    1	        1
    A2,	                any 	                1	        any
    A3	                4 or 5	                any     	any
    A4	                any 	                1	        any
    A5	                4 or 5	                any     	any
    B1,B2	            2 or 3	                any	        any
    N0	                any except 4 and 5	    1	        any
    N1	                any except 4 and 5	    1	        any
    LEGACY500	        any except 4 and 5 	    1	        any
    T1	                4 or 5	                any     	any
    F3	                any except 4 and 5 	    1	        any
    F4	                4 or 5	                any	        any
    F5	                4 or 5	                any	        any
    F6	                4 or 5	                any	        any
    F7	                4 or 5	                any	        any
    KC,KT,KS	        any except 4 and 5 	    any	        any

    default packets
    ID,T0,F1, F2,S1, VR,VA
    '''
    packet_types = []
    default_packet_types = ['S1']  # ['ID','T0','F1','F2','S1','VR','VA']

    packet_types.extend(default_packet_types)

    if architecture != 4 and architecture != 5:
        # packet_types.extend(['KC', 'KT', 'KS'])
        if mags == 1:
            packet_types.append('S0')
        if algorithm == 1:
            packet_types.extend(['S2', 'N0', 'N1'])
        if mags == 1 and algorithm == 1:
            packet_types.extend(['A0', 'A1', 'LEGACY500', 'F3'])

    if architecture == 4 or architecture == 5:
        packet_types.extend(['S3', 'A3', 'A5', 'T1', 'F4', 'F5', 'F6', 'F7'])

    if architecture == 2 or architecture == 3:
        packet_types.extend(['B1', 'B2'])

    if algorithm == 1:
        packet_types.extend(['A2', 'A4'])

    packet_types.sort()
    return packet_types



from .utils.resource import is_dev_mode
from .utils.helper import dict_to_object

DEV_CONFIG = dict_to_object({
    'ANS_PLATFORM_URL': 'http://localhost:3000/',
    'AZURE_STORAGE_ACCOUNT': 'navview',
    'AZURE_STORAGE_BACKUP_CONTAINER': 'testing',
    'AZURE_STORAGE_DATA_CONTAINER': 'data-1000',
    'AZURE_STORAGE_APPS_CONTAINER': 'apps'
})

PROD_CONFIG = dict_to_object({
    'ANS_PLATFORM_URL': 'https://api.aceinna.com/',
    'AZURE_STORAGE_ACCOUNT': 'navview',
    'AZURE_STORAGE_BACKUP_CONTAINER': 'backup',
    'AZURE_STORAGE_DATA_CONTAINER': 'data',
    'AZURE_STORAGE_APPS_CONTAINER': 'apps'
})


def get_config():
    '''
    Get configuration by dev mode
    '''
    if is_dev_mode():
        return DEV_CONFIG

    return PROD_CONFIG



import functools
from .message_center import (DeviceMessage)


def with_device_message(func):
    '''
    This is a decorator for method with DeviceMessage, it would looks like
    code: yield message_center.build(command=command_line)
    '''

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        generator_func = func(*args, **kwargs)
        global generator_result
        generator_result = None

        def check_result():
            global generator_result
            while not generator_result:
                continue

            if generator_result:
                next_device_message = generator_func.send(generator_result)
                if isinstance(next_device_message, DeviceMessage):
                    generator_result = None
                    next_device_message.on('finished', on_resolve)
                    next_device_message.send()
                    return check_result()
                else:
                    return next_device_message

        def on_resolve(packet_type, data, error):
            global generator_result
            generator_result = {
                'packet_type': packet_type,
                'data': data,
                'error': error
            }

        try:
            device_message = generator_func.send(None)
            if isinstance(device_message, DeviceMessage):
                device_message.on('finished', on_resolve)
                device_message.send()
                return check_result()
            else:
                return device_message
        except StopIteration as ex:
            value = {
                'packetType': 'error',
                'data': 'No Response'
            }

            if hasattr(ex, 'value'):
                value = ex.value
            return value

    return wrapper



import json
import requests
from .configuration import get_config


class AnsPlatformAPI:
    def __init__(self):
        self.access_token = ''
        self.host_url = get_config().ANS_PLATFORM_URL

    def set_access_token(self, token):
        self.access_token = token

    def get_sas_token(self):
        try:
            url = self.host_url + "token/storagesas"
            headers = {'Content-type': 'application/json',
                       'Authorization': self.access_token}
            response = requests.post(url, headers=headers)
            rev = response.json()
            if 'token' in rev:
                return rev['token']
            else:
                return ''
        except Exception as ex:
            print('Exception when get_sas_token:', ex)
            return ''

    def save_backup_restult(self, serial_num, file_name, device_type):
        ''' save backup result
        '''
        body = {
            'data': {
                'sn': serial_num,
                'file': file_name,
                'type': device_type
            }
        }
        try:
            url = self.host_url + "api/userDevices/backup"
            data_json = json.dumps(body)
            headers = {'Content-type': 'application/json',
                       'Authorization': self.access_token}
            response = requests.post(url, data=data_json, headers=headers)
            return response.json()
        except Exception as ex:
            print('Exception when update db:', ex)



from .parsers.open_message_parser import UartMessageParser as OpenUartMessageParser
from .parsers.dmu_message_parser import UartMessageParser as DMUUartMessageParser


class ParserManager:
    '''
    Manage Parser
    '''
    device_list = []

    @staticmethod
    def build(device_type, properties):
        '''
        Generate matched parser
        '''
        if device_type == 'DMU':
            return DMUUartMessageParser(properties)
        else:
            return OpenUartMessageParser(properties)



import struct
import collections
from ..dmu.configuration_field import CONFIGURATION_FIELD_DEFINES_SINGLETON
from ..dmu.eeprom_field import EEPROM_FIELD_DEFINES_SINGLETON
from .dmu_field_parser import decode_value
# input packet


def read_eeprom_parser(payload):
    read_address = decode_value('uint16', payload[0:2])
    eeprom_data = payload[3:]
    eeprom_field = EEPROM_FIELD_DEFINES_SINGLETON.find(read_address)
    value, parsed, error = eeprom_field.parse(eeprom_data)

    return {
        "name": eeprom_field.name,
        "value": value,
        "parsed": parsed
    }, error


def read_field_parser(payload):
    data = []
    error = False
    number_of_fields = payload[0]
    data_payload = payload[1:]

    for parameter_index in range(number_of_fields):
        parameter_id = decode_value(
            'uint16', data_payload[parameter_index * 4:parameter_index*4+2])

        parameter_value = data_payload[parameter_index *
                                       4+2:parameter_index*4+4]

        configuration_field = CONFIGURATION_FIELD_DEFINES_SINGLETON.find(
            parameter_id)

        value, parsed, error = configuration_field.parse(parameter_value)
        # value = unpack_value(configuration_field.field_type, parameter_value)

        if error:
            break

        data.append({
            "paramId": parameter_id,
            "name": configuration_field.name,
            "value": value,
            "parsed": parsed
        })

    return data, error


def write_field_parser(payload):
    data = 0
    error = False
    field_count = decode_value('uint8', payload[0:1])

    if isinstance(field_count, bool) and not field_count:
        error = True
        data = -1

    return data, error


# output packet

def id_parser(payload, *args):
    '''
    Parse id packet
    '''
    serial_num = int.from_bytes(struct.pack(
        '4B', *payload[0:4]), byteorder='big')

    mode_string_len = len(payload[4:])
    model_string = struct.pack('{0}B'.format(
        mode_string_len), *payload[4:]).decode()

    split_text = model_string.split(' ')

    return {
        'name': split_text[0],
        'pn': split_text[1],
        'sn': serial_num
    }


def version_data_parser(payload, *args):
    '''
    Parse version data
    '''
    version_string = '{0}.{1}.{2}.{3}.{4}'.format(*payload)

    return {
        'app_name': 'DMU',
        'version': version_string
    }


def common_continuous_parser(payload, configuration, scaling):
    '''
    Unpack output packet
    '''
    if configuration is None:
        return

    data = None
    length = 0
    pack_fmt = '>'
    for value in configuration['payload']:
        if value['type'] == 'float':
            pack_fmt += 'f'
            length += 4
        elif value['type'] == 'uint32':
            pack_fmt += 'I'
            length += 4
        elif value['type'] == 'int32':
            pack_fmt += 'i'
            length += 4
        elif value['type'] == 'int16':
            pack_fmt += 'h'
            length += 2
        elif value['type'] == 'uint16':
            pack_fmt += 'H'
            length += 2
        elif value['type'] == 'double':
            pack_fmt += 'd'
            length += 8
        elif value['type'] == 'int64':
            pack_fmt += 'q'
            length += 8
        elif value['type'] == 'uint64':
            pack_fmt += 'Q'
            length += 8
        elif value['type'] == 'char':
            pack_fmt += 'c'
            length += 1
        elif value['type'] == 'uchar':
            pack_fmt += 'B'
            length += 1
        elif value['type'] == 'uint8':
            pack_fmt += 'B'
            length += 1
    len_fmt = '{0}B'.format(length)

    try:
        pack_item = struct.pack(len_fmt, *payload)
        data = struct.unpack(pack_fmt, pack_item)
        out = []

        for idx, item in enumerate(configuration['payload']):
            scaling_setting = None
            scaling_value = 1
            if item.__contains__('scaling'):
                scaling_setting = scaling[item['scaling']]
            if scaling_setting:
                scaling_value = eval(scaling_setting)

            format_value = data[idx]*scaling_value
            out.append((item['name'], format_value))

        format_data = collections.OrderedDict(out)

    except Exception as ex:  # pylint: disable=broad-except
        print(
            "error happened when decode the payload of packets, pls restart driver: {0}"
            .format(ex))

    return format_data


# packet handler
def match_command_handler(packet_type):
    '''
    Find the handler for specified packet
    '''
    parser_dict = {
        'RF': read_field_parser,
        'GF': read_field_parser,
        'SF': write_field_parser,
        'WF': write_field_parser,
        'RE': read_eeprom_parser
    }
    return parser_dict.get(packet_type)


def match_continuous_handler(packet_type):
    parser_dict = {
        'ID': id_parser,
        'VR': version_data_parser
    }
    if not parser_dict.__contains__(packet_type):
        return common_continuous_parser
    return parser_dict.get(packet_type)



"""
Websocket server entry
"""
import os
import sys

import json
import traceback
import threading
import tornado.websocket
import tornado.ioloop
import tornado.httpserver
import tornado.web
from tornado import gen
from .. import VERSION
from ..devices.base.event_base import EventBase
from ..framework.communicator import CommunicatorFactory
from ..framework.context import APP_CONTEXT
from ..framework.file_storage import FileLoger
from ..framework.utils import (helper, resource)
from ..models import WebserverArgs
from ..framework.constants import DEFAULT_PORT_RANGE
from ..framework import AppLogger
if sys.version_info[0] > 2:
    from queue import Queue
else:
    from Queue import Queue

SERVER_UPDATE_RATE = 50


class WSHandler(tornado.websocket.WebSocketHandler):
    '''
    Websocket handler
    '''
    is_streaming = False
    is_logging = False
    latest_packet_collection = []
    file_logger = None
    packet_white_list = ['ping', 'upgrade_progress',
                         'upgrade_complete', 'mag_status', 'backup_status', 'restore_status']
    period_output_callback = None

    def initialize(self, server):
        '''
        Websocket handler initialize
        '''
        server.ws_handler = self

    def open(self):
        connected_device = self.get_device()
        if connected_device:
            self.handle_device_found(connected_device)
        else:
            self.response_device_isnot_connected()

        self.period_output_callback = tornado.ioloop.PeriodicCallback(
            self.response_output_packet, SERVER_UPDATE_RATE)
        self.period_output_callback.start()

    def on_message(self, message):
        client_msg = json.loads(message)
        method = client_msg['method'] if 'method' in client_msg else None
        parameters = client_msg['params'] if 'params' in client_msg else None
        if method:
            try:
                self.handle_message(method, parameters)
            except Exception as ex:  # pylint:disable=broad-except
                print('websocket on message error', ex)
                traceback.print_exc()
                self.response_message(
                    method, {'packetType': 'error', 'data': 'sever error'})
        else:
            self.response_unkonwn_method()

    def on_close(self):
        self.reset()
        try:
            self.period_output_callback.stop()
        except:  # pylint:disable=bare-except
            # need log exception
            pass
        connected_device = self.get_device()
        if connected_device:
            connected_device.remove_client(self)
            # print('close client count:', len(connected_device.clients))

    def check_origin(self, origin):
        return True

    def get_device(self):
        '''
        Get device
        '''
        return APP_CONTEXT.get_app().get_device()

    def on_receive_output_packet(self, method, packet_type, data):
        '''
        Listenr for receive output packet
        '''
        data_updated = False
        for item in self.latest_packet_collection:
            if item['packet_type'] == packet_type:
                item['data'] = data
                data_updated = True
                break

        if not data_updated:
            self.latest_packet_collection.append({
                'packet_type': packet_type,
                'data': data
            })

        if self.file_logger and self.is_logging:
            self.file_logger.append(packet_type, data)

    def reset(self):
        '''
        Reset status
        '''
        self.is_streaming = False
        self.is_logging = False
        if self.file_logger:
            self.file_logger.stop_user_log()

    def handle_message(self, method, parameters):
        '''
        Handle received message
        '''
        device = self.get_device()

        converted_method = helper.name_convert_camel_to_snake(method)

        if device and device.connected and hasattr(device, converted_method):
            device_command = getattr(
                device, converted_method, None)(parameters)
            self.response_message(method, device_command)

        elif hasattr(self, converted_method):
            getattr(self, converted_method, None)(parameters)
        else:
            self.response_unkonwn_method()

    def handle_device_found(self, device, force_response=True):
        '''
        If detect device, setup output and logger
        '''
        if len(device.clients) == 1:
            self.response_only_allow_one_client()
            return

        device.append_client(self)
        # print('open client count:', len(device.clients))

        self.file_logger = FileLoger(device.properties)
        if force_response:
            self.response_server_info(device)

    def response_message(self, method, data):
        '''
        Format response
        '''
        self.write_message(
            json.dumps(
                {
                    'method': method,
                    'result': data
                }
            ))

    def response_unkonwn_method(self):
        '''
        Format unknonwn method message
        '''
        self.write_message(
            json.dumps({
                'method': 'unknown',
                'result': {
                    'packetType': 'error',
                    'data': {
                        'message': 'unknown method'
                    }
                }
            })
        )

    def response_server_info(self, device):
        '''
        Send webserver info
        '''
        self.response_message('stream', {
            'packetType': 'serverInfo',
            'data': {
                'version': VERSION,
                'serverUpdateRate': 50,
                'deviceConnected': True,
                'clientCount': 0,
                'deviceType': device.type
            }})

    def response_output_packet(self):
        '''
        Response streaming data
        '''
        for latest_packet in self.latest_packet_collection:
            if latest_packet['packet_type'] in self.packet_white_list or self.is_streaming:
                self.write_message(
                    json.dumps({
                        'method': 'stream',
                        'result': {
                            'packetType': latest_packet['packet_type'],
                            'data': latest_packet['data']
                        }
                    })
                )

        helper.clear_elements(self.latest_packet_collection)

    def response_device_isnot_connected(self):
        '''
        Response device is not connected
        '''
        self.response_message('stream', {
            'packetType': 'serverInfo',
            'data': {
                'version': VERSION,
                'serverUpdateRate': SERVER_UPDATE_RATE,
                'deviceConnected': False,
                'clientCount': 0
            }})

    def response_only_allow_one_client(self):
        '''
        Response only allow one client connect
        '''
        self.response_message('stream', {
            'packetType': 'serverInfo',
            'data': {
                'version': VERSION,
                'serverUpdateRate': SERVER_UPDATE_RATE,
                'deviceConnected': True,
                'clientCount': 1
            }})
    # protocol

    def start_stream(self, *args):  # pylint: disable=invalid-name
        '''
        Start to send stream data
        '''
        self.response_message('startStream', {'packetType': 'success'})
        self.is_streaming = True

    def stop_stream(self, *args):  # pylint: disable=invalid-name
        '''
        Stop sending stream data
        '''
        self.response_message('stopStream', {'packetType': 'success'})
        self.is_streaming = False

    def start_log(self, *args):  # pylint: disable=invalid-name
        '''
        Start record log
        '''
        parameters = args[0]
        self.file_logger.set_info(self.get_device().get_log_info())
        self.file_logger.set_user_id(parameters['id'])
        self.file_logger.set_user_access_token(parameters['access_token'])
        self.file_logger.start_user_log(parameters['fileName'], True)
        self.is_logging = True
        self.response_message(
            'startLog', {'packetType': 'success', 'data': parameters['fileName']+'.csv'})

    def stop_log(self, *args):  # pylint: disable=invalid-name
        '''
        Stop record log
        '''
        self.file_logger.stop_user_log()
        self.is_logging = False
        self.response_message('stopLog', {'packetType': 'success', 'data': ''})


class MessageStore(object):
    def __init__(self):
        self.messages = Queue()
        self.ready = False
        self.max_length = 100

    def append(self, msg):
        format_msg = '{} - {}: {}'.format(
            msg['asctime'],
            msg['levelname'],
            msg['message']
        )

        current_msg_len = self.messages.qsize()
        overflow_len = current_msg_len - self.max_length + 1
        for _ in range(overflow_len):
            self.messages.get()

        self.messages.put(format_msg)

    def get_messages(self):
        output_msg = []
        current_msg_len = self.messages.qsize()
        for _ in range(current_msg_len):
            output_msg.append(self.messages.get())
        return output_msg

    def empty(self):
        self.messages.empty()

    def size(self):
        return self.messages.qsize()


class LoggerServerSentEvent(tornado.web.RequestHandler):
    '''
    Send device data to client
    '''

    def __init__(self, *args, **kwargs):
        super(LoggerServerSentEvent, self).__init__(*args, **kwargs)
        self.set_header('Content-Type', 'text/event-stream')
        self.set_header('Access-Control-Allow-Origin', '*')

    def initialize(self, store):
        '''
        Websocket handler initialize
        '''
        self.store = store
        self._auto_finish = False

    def get(self):
        self._loop = tornado.ioloop.PeriodicCallback(
            lambda: {self.emit()}, 1000)
        self._loop.start()

    @gen.coroutine
    def emit(self):
        '''
        Fetch log information from logger
        '''
        try:
            if self.store.size() == 0:
                yield self.flush()

            self.write('data:{}\n\n'.format(
                json.dumps({
                    'msg': self.store.get_messages()
                })
            ))
            yield self.flush()
        except tornado.iostream.StreamClosedError as e:
            self._loop.stop()
            self.finish()
        except RuntimeError as e:
            self._loop.stop()
            self.finish()


class Webserver(EventBase):
    '''
    Websocket server
    '''

    def __init__(self, **kwargs):
        super(Webserver, self).__init__()
        self.communication = 'uart'
        self.device_provider = None
        self.communicator = None
        self.ws_handler = None
        self.sse_handler = None
        self.http_server = None
        self._build_options(**kwargs)
        APP_CONTEXT.set_app(self)

        self.prepare_logger()

    def listen(self):
        '''
        Start to find device
        '''
        # start websocket server
        webserver_thread = threading.Thread(target=self.start_webserver)
        webserver_thread.start()

        self.detect_device(self.device_discover_handler)

    def prepare_logger(self):
        '''
        Set default log handler: console logger, file logger
        '''
        executor_path = resource.get_executor_path()
        log_level = 'info'
        if self.options.debug:
            log_level = 'debug'

        console_log = self.options.console_log

        APP_CONTEXT.set_logger(
            AppLogger(
                filename=os.path.join(executor_path, 'loggers', 'trace.log'),
                gen_file=True,
                level=log_level,
                console_log=console_log
            ))

    def get_device(self):
        '''
        Get device provider
        '''
        if self.device_provider is not None:
            return self.device_provider
        return None

    def device_discover_handler(self, device_provider):
        '''
        Handler after device discovered
        '''
        # load device provider
        self.load_device_provider(device_provider)
        # notify ws handler
        if self.ws_handler:
            self.ws_handler.handle_device_found(
                device_provider)

    def device_rediscover_handler(self, device_provider):
        '''
        Handler after device rediscovered
        '''
        if self.device_provider.device_info['sn'] == device_provider.device_info['sn']:
            if self.ws_handler:
                self.ws_handler.on_receive_output_packet(
                    'stream', 'ping', {'status': 3})
            self.load_device_provider(device_provider)
        else:
            self.device_provider.close()
            self.load_device_provider(device_provider)
            if self.ws_handler:
                self.ws_handler.on_receive_output_packet(
                    'stream', 'ping', {'status': 1})
                self.ws_handler.handle_device_found(device_provider, False)


    def device_complete_upgrade_handler(self, device_provider):
        '''
        Handler after device upgrade complete
        '''
        self.device_provider = device_provider
        self.device_provider.upgrade_completed(self.options)
        if self.device_provider.device_info['sn'] == device_provider.device_info['sn']:
            if self.ws_handler:
                self.ws_handler.on_receive_output_packet(
                    'stream', 'upgrade_complete', {'success': True})
        else:
            if self.ws_handler:
                self.ws_handler.on_receive_output_packet(
                    'stream', 'upgrade_complete', {'success': False})
         #   self.device_provider.close()

    def load_device_provider(self, device_provider):
        '''
        Load device provider
        '''
        self.device_provider = device_provider
        self.device_provider.setup(self.options)
        self.device_provider.on('exception', self.handle_device_exception)
        self.device_provider.on(
            'complete_upgrade', self.handle_device_complete_upgrade)

    def start_webserver(self):
        # self.webserver_io_loop = asyncio.new_event_loop()
        if sys.version_info[0] > 2:
            import asyncio
            asyncio.set_event_loop(asyncio.new_event_loop())
        self.start_websocket_server()

    def start_websocket_server(self):
        '''
        Initial websocket server
        '''
        # add ws handler as a member
        store = MessageStore()
        try:
            application = tornado.web.Application(
                [
                    (r'/', WSHandler, dict(server=self)),
                    (r'/sse', LoggerServerSentEvent, dict(store=store))
                ])
            self.http_server = tornado.httpserver.HTTPServer(application)
            # self.http_server.listen(self.options.port)
            activated_port = 0
            if self.options.port == 'auto':
                for webserver_port in DEFAULT_PORT_RANGE:
                    try:
                        self.http_server.listen(webserver_port)
                        activated_port = webserver_port
                        break
                    except:
                        continue
                if activated_port == 0:
                    raise Exception('Port in used')
            else:
                self.http_server.listen(self.options.port)
                activated_port = self.options.port
            print('Websocket server is started on port', activated_port)

            APP_CONTEXT.get_logger().enable_msg_store_handler(store)
            tornado.ioloop.IOLoop.instance().start()
        except Exception as ex:
            print(ex)
            # print('Cannot start a websocket server, please check if the port is in use')
            raise

    def handle_device_exception(self, error, message):
        '''
        Handle device exception
        '''
        if self.ws_handler:
            self.ws_handler.reset()
            self.ws_handler.on_receive_output_packet(
                'stream', 'ping', {'status': 2})

        self.device_provider.reset()
        self.detect_device(self.device_rediscover_handler)

    def handle_device_complete_upgrade(self):
        '''
        Handle device complete upgrade
        '''
        self.communicator.reset_buffer()
        self.communicator.close()
        # self.device_provider.reset()
        self.detect_device(self.device_complete_upgrade_handler)

    def detect_device(self, callback):
        '''find if there is a connected device'''
        print('Prepare to find device...')
        if self.communicator is None:
            self.communicator = CommunicatorFactory.create(
                self.communication, self.options)

        self.communicator.find_device(callback)

    def set_communicator(self, communicator):
        self.communicator = communicator

    def stop(self):
        '''close websocket server'''
        if self.http_server is not None:
            self.http_server.stop()

    def _build_options(self, **kwargs):
        self.options = WebserverArgs(**kwargs)



#import asyncio
import sys
import threading
import datetime
import time
from .base.event_base import EventBase
from ..framework.utils import helper
if sys.version_info[0] > 2:
    from queue import Queue
else:
    from Queue import Queue


class EventMessage(EventBase):
    def __init__(self):
        super(EventMessage, self).__init__()

    def set_result(self, result):
        self.emit('finished', result)

    def then(self, callback):
        self.on('finished', callback)


class DeviceMessage(EventBase):
    def __init__(self, message_center, command, timeout=1):
        super(DeviceMessage, self).__init__()
        self.message_id = None
        self.result = None
        self._message_center = message_center
        self._command = command
        # self._packet_type = packet_type
        self._status = ''
        self._start_time = None
        self._timeout = timeout
        self._is_finished = False

    def send(self):
        self._message_center.append_and_run(self)

    def finish(self, **kwargs):
        if not self._is_finished:
            self._is_finished = True
            self.emit('finished', **kwargs)

    def set_status(self, status):
        self._status = status

    def set_start_time(self, start_time):
        self._start_time = start_time

    def get_start_time(self):
        return self._start_time

    def get_command(self):
        return self._command

    def get_timeout(self):
        return self._timeout

    def get_finished(self):
        return self._is_finished


class DeviceMessageCenter(EventBase):
    '''
    Device message center, it handles status of message, and also work as a message factory
    '''

    def __init__(self, communicator):
        super(DeviceMessageCenter, self).__init__()
        self.threads = []
        self._communicator = communicator
        self._is_stop = False
        self._is_pause = False
        self._has_exception = False
        self.data_queue = Queue()  # data container
        self.data_lock = threading.Lock()
        self.exception_lock = threading.Lock()
        self._is_running = False
        self.prerun_queue = Queue()
        self._parser = None
        self._running_message = None
        self._is_ready = False
        self._has_running_checker = False

    def is_ready(self):
        '''Check if message center is setuped
        '''
        return self._is_ready

    def set_parser(self, parser):
        self._parser = parser
        self._parser.on('command', self.on_command_receive)
        self._parser.on('continuous_message',
                        self.on_continuous_messageReceive)

    def get_parser(self):
        return self._parser

    def build(self, command, timeout=1):
        return DeviceMessage(self, command, timeout)

    def append_and_run(self, message):
        # self._msg_list.append(message)
        self.request_run(message)

    def request_run(self, message):
        if self._is_running:
            self.prerun_queue.put(message)
        else:
            self.run(message)

    def run(self, message):
        self._is_running = True
        self._running_message = message
        message.set_start_time(datetime.datetime.now())
        
        self._parser.set_run_command(message.get_command())
        self._communicator.write(message.get_command())
        #print('run command', message.get_command())

    def run_post(self):
        if self.prerun_queue.empty():
            self._is_running = False
        else:
            next_message = self.prerun_queue.get()
            self.run(next_message)
        # print('post')

    def setup(self):
        if not self._has_running_checker:
            thread = threading.Thread(
                target=self.thread_running_checker, args=())
            thread.start()
            self._has_running_checker = True

        # setup receiver, parser
        funcs = [self.thread_receiver, self.thread_parser]
        for func in funcs:
            thread = threading.Thread(target=func, args=())
            thread.start()
            self.threads.append(thread)

        self._is_ready = True

    def pause(self):
        self._is_pause = True

    def resume(self):
        self._has_exception = False
        self._is_pause = False
        helper.clear_elements(self.threads)

    def stop(self):
        self._is_stop = True

    def timeout_check(self):
        if self._is_running:
            timeout = self._running_message.get_timeout()
            start_time = self._running_message.get_start_time()
            current_time = datetime.datetime.now()

            span = current_time - start_time
            if span.total_seconds() > timeout and not \
                    self._running_message.get_finished():
                self.run_post()
                print('command timeout',
                      self._running_message.get_command(),
                      timeout, start_time, current_time)
                self._running_message.finish(
                    error='Timeout', packet_type=None, data=None)

    def thread_running_checker(self):
        '''
        Check running status
        '''
        if sys.version_info[0] > 2:
            import asyncio
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        while True:
            self.exception_lock.acquire()
            if self._has_exception:
                # self.connected = False
                self.emit('error', 'app', 'communicator read error')
            self.exception_lock.release()

            if self._is_stop:
                return

            self.timeout_check()

            try:
                time.sleep(0.1)
            except KeyboardInterrupt:  # response for KeyboardInterrupt such as Ctrl+C
                return True

    def thread_receiver(self, *args, **kwargs):
        ''' receive rover data and push data into data_queue.
            return when occur Exception
        '''
        while True:
            if self._has_exception or self._is_stop:
                return

            if self._is_pause:
                time.sleep(0.1)
                continue

            data = None
            try:
                data = self._communicator.read()
            except Exception as ex:  # pylint: disable=broad-except
                print('Thread:receiver error:', ex)
                self.exception_lock.acquire()
                self._has_exception = True  # Notice thread paser to exit.
                self.exception_lock.release()
                return  # exit thread receiver

            if data and len(data) > 0:
                self.emit('read_block', data)
                self.data_lock.acquire()
                for data_byte in data:
                    self.data_queue.put(data_byte)
                self.data_lock.release()
            else:
                time.sleep(0.001)

    def thread_parser(self, *args, **kwargs):
        ''' get rover data from data_queue and parse data into one whole frame.
            return when occur Exception in thread receiver.
        '''
        if sys.version_info[0] > 2:
            import asyncio
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        while True:
            if self._has_exception or self._is_stop:
                return

            self.exception_lock.acquire()
            if self._has_exception:
                self.exception_lock.release()
                return  # exit thread parser
            self.exception_lock.release()

            if self._is_pause:
                time.sleep(0.1)
                continue

            self.data_lock.acquire()
            if self.data_queue.empty():
                self.data_lock.release()
                time.sleep(0.001)
                continue
            else:
                data = self.data_queue.get()
                self.data_lock.release()

            if self._parser:
                if sys.version_info[0] < 3:
                    data = ord(data)
                self._parser.analyse(data)

    def on_command_receive(self, *args, **kwargs):
        self.run_post()
        if self._running_message:
            self._running_message.finish(**kwargs)

    def on_continuous_messageReceive(self, *args, **kwargs):
        self.emit('continuous_message', **kwargs)



"""
predefined params for openrtk
"""
JSON_FILE_NAME = 'openrtk.json'

def get_app_names():
    '''
    define openimu app type
    '''
    app_names = ['INS',
                 'RAWDATA',
                 'RTK',
                 ]
    return app_names


APP_STR = ['INS', 'RAWDATA', 'RTK']



import os
import time
import json
import binascii
import math
# import asyncio
import datetime
import threading
import struct
from azure.storage.blob import BlockBlobService
from ...framework.utils import helper
from ...framework.utils import resource
from ..base.uart_base import OpenDeviceBase
from ..configs.openimu_predefine import (
    APP_STR, get_app_names
)
from ...framework.context import APP_CONTEXT
from ..decorator import with_device_message
from ...framework.ans_platform_api import AnsPlatformAPI
from ...framework.configuration import get_config


class Provider(OpenDeviceBase):
    '''
    OpenIMU UART provider
    '''

    def __init__(self, communicator):
        super(Provider, self).__init__(communicator)
        self.type = 'IMU'
        self.server_update_rate = 50
        self.is_logging = False
        self.is_mag_align = False
        self.bootloader_baudrate = 57600
        self.device_info = None
        self.app_info = None
        self.app_config_folder = ''
        self.parameters = None
        self.enable_data_log = True
        self.prepare_folders()
        self.is_backup = False
        self.is_restore = False
        self.is_app_matched = False
        self.ans_platform = AnsPlatformAPI()

    def prepare_folders(self):
        '''
        Prepare folders for data storage and configuration
        '''
        executor_path = resource.get_executor_path()
        setting_folder_name = 'setting'
        config_file_name = 'openimu.json'
        data_folder_path = os.path.join(executor_path, 'data')
        if not os.path.isdir(data_folder_path):
            os.makedirs(data_folder_path)

        # copy contents of app_config under executor path
        self.setting_folder_path = os.path.join(
            executor_path, setting_folder_name, 'openimu')

        for app_name in get_app_names():
            app_name_path = os.path.join(self.setting_folder_path, app_name)
            app_name_config_path = os.path.join(
                app_name_path, config_file_name)
            if not os.path.isfile(app_name_config_path):
                if not os.path.isdir(app_name_path):
                    os.makedirs(app_name_path)
                app_config_content = resource.get_content_from_bundle(
                    setting_folder_name, os.path.join('openimu', app_name, config_file_name))
                if app_config_content is None:
                    continue

                with open(app_name_config_path, "wb") as code:
                    code.write(app_config_content)

    def ping(self):
        '''
        Check if the connected device is OpenIMU
        '''
        # print('start to check if it is openimu')
        device_info_text = self.internal_input_command('pG')
        app_info_text = self.internal_input_command('gV')

        # TODO: Prevent action. Get app info again,
        # if cannot retrieve any info at the first time of ping. Should find the root cause.
        if app_info_text == '':
            app_info_text = self.internal_input_command('gV')

        APP_CONTEXT.get_logger().logger.debug('Checking if is OpenIMU device...')
        APP_CONTEXT.get_logger().logger.debug(
            'Device: {0}'.format(device_info_text))
        APP_CONTEXT.get_logger().logger.debug(
            'Firmware: {0}'.format(app_info_text))

        if device_info_text.find('OpenIMU') > -1 and \
                device_info_text.find('OpenRTK') == -1:
            self.build_device_info(device_info_text)
            self.build_app_info(app_info_text)
            self.connected = True
            print('# Connected Information #')
            print('Device:', device_info_text)
            print('Firmware:', app_info_text)
            APP_CONTEXT.get_logger().logger.info(
                'Connected {0}, {1}'.format(device_info_text, app_info_text))
            return True
        return False

    def build_device_info(self, text):
        '''
        Build device info
        '''
        split_text = text.split(' ')
        split_len = len(split_text)
        pre_sn = split_text[3].split(':') if split_len == 4 else ''
        serial_num = pre_sn[1] if len(pre_sn) == 2 else ''
        self.device_info = {
            'name': split_text[0],
            'pn': split_text[1],
            'firmware_version': split_text[2],
            'sn': serial_num
        }

    def build_app_info(self, text):
        '''
        Build app info
        '''
        # check if J1939 in text
        app_version = text
        can_indicator = '_J1939'
        if can_indicator in app_version:
            app_version = app_version.replace(can_indicator, '')

        split_text = app_version.split(' ')
        app_name = next(
            (item for item in APP_STR if item in split_text), None)

        if not app_name:
            app_name = 'IMU'
            self.is_app_matched = False
        else:
            self.is_app_matched = True

        self.app_info = {
            'app_name': app_name,
            'version': text
        }

    def load_properties(self):
        # Load config from user working path
        local_config_file_path = os.path.join(os.getcwd(), 'openimu.json')
        if os.path.isfile(local_config_file_path):
            with open(local_config_file_path) as json_data:
                self.properties = json.load(json_data)
                return

        # Load the openimu.json based on its app
        app_name = self.app_info['app_name']
        app_file_path = os.path.join(
            self.setting_folder_path, app_name, 'openimu.json')

        if not self.is_app_matched:
            APP_CONTEXT.get_logger().warning(
                ('Failed to extract app version information from unit. The supported application list is {0}.').format(get_app_names()))
            APP_CONTEXT.get_logger().warning(
                'To keep runing, use IMU configuration as default.')
            APP_CONTEXT.get_logger().warning(
                'You can choose to place your json file under exection path if it is an unknown application.')

        with open(app_file_path) as json_data:
            self.properties = json.load(json_data)

    def after_setup(self):
        pass

    def on_read_raw(self, data):
        pass

    def on_receive_output_packet(self, packet_type, data):
        '''
        Listener for getting output packet
        '''
        self.add_output_packet('stream', packet_type, data)

    def get_log_info(self):
        '''
        Build information for log
        '''
        packet_rate = next(
            (item['value'] for item in self.parameters if item['name'] == 'Packet Rate'), '100')
        return {
            "type": self.type,
            "model": self.device_info['name'],
            "logInfo": {
                "pn": self.device_info['pn'],
                "sn": self.device_info['sn'],
                "sampleRate": packet_rate,
                "appVersion": self.app_info['version'],
                "imuProperties": json.dumps(self.properties)
            }
        }

    # command list
    def get_device_info(self, *args):  # pylint: disable=invalid-name
        '''
        Get device information
        '''
        return {
            'packetType': 'deviceInfo',
            'data':  [
                {'name': 'Product Name',
                 'value': self.device_info['name']},
                {'name': 'PN', 'value': self.device_info['pn']},
                {'name': 'Firmware Version',
                 'value': self.device_info['firmware_version']},
                {'name': 'SN', 'value': self.device_info['sn']},
                {'name': 'App Version', 'value': self.app_info['version']}
            ]
        }

    def get_conf(self, *args):  # pylint: disable=invalid-name
        '''
        Get json configuration
        '''
        return {
            'packetType': 'conf',
            'data': {
                'outputs': self.properties['userMessages']['outputPackets'],
                'inputParams': self.properties['userConfiguration']
            }
        }

    @with_device_message
    def get_params(self, *args):  # pylint: disable=invalid-name
        '''
        Get all parameters
        '''
        command_line = helper.build_input_packet('gA')

        result = yield self._message_center.build(command=command_line, timeout=3)

        data = result['data']

        if data:
            self.parameters = data
            yield {
                'packetType': 'inputParams',
                'data': data
            }
        yield {
            'packetType': 'error',
            'data': 'No Response'
        }

    @with_device_message
    def get_param(self, params, *args):  # pylint: disable=invalid-name
        '''
        Update paramter value
        '''
        command_line = helper.build_input_packet(
            'gP', properties=self.properties, param=params['paramId'])

        result = yield self._message_center.build(command=command_line)

        data = result['data']
        if data:
            yield {
                'packetType': 'inputParam',
                'data': data
            }
        yield {
            'packetType': 'error',
            'data': 'No Response'
        }

    @with_device_message
    def set_params(self, params, *args):  # pylint: disable=invalid-name
        '''
        Update paramters value
        '''
        for parameter in params:
            command_line = helper.build_input_packet(
                'uP', properties=self.properties,
                param=parameter['paramId'],
                value=parameter['value'])

            result = yield self._message_center.build(command=command_line)

            packet_type = result['packet_type']
            data = result['data']

            if packet_type == 'error':
                yield {
                    'packetType': 'error',
                    'data': {
                        'error': data
                    }
                }
            if data > 0:
                yield {
                    'packetType': 'error',
                    'data': {
                        'error': data
                    }
                }

        yield {
            'packetType': 'success',
            'data': {
                'error': 0
            }
        }

    @with_device_message
    def set_param(self, params, *args):  # pylint: disable=invalid-name
        '''
        Update paramter value
        '''
        command_line = helper.build_input_packet(
            'uP', properties=self.properties, param=params['paramId'], value=params['value'])

        result = yield self._message_center.build(command=command_line)

        error = result['error']
        data = result['data']
        if error:
            yield {
                'packetType': 'error',
                'data': {
                    'error': data
                }
            }

        yield {
            'packetType': 'success',
            'data': {
                'error': data
            }
        }

    @with_device_message
    def reset_params(self, params, *args):  # pylint: disable=unused-argument
        '''
        Reset params to default
        '''
        command_line = helper.build_input_packet('rD')
        result = yield self._message_center.build(command=command_line, timeout=2)

        error = result['error']
        data = result['data']
        if error:
            yield {
                'packetType': 'error',
                'data': {
                    'error': error
                }
            }

        yield {
            'packetType': 'success',
            'data': data
        }

    @with_device_message
    def save_config(self, *args):  # pylint: disable=unused-argument
        '''
        Save configuration
        '''
        command_line = helper.build_input_packet('sC')
        result = yield self._message_center.build(command=command_line)

        data = result['data']
        error = result['error']
        if data:
            yield {
                'packetType': 'success',
                'data': data
            }

        yield {
            'packetType': 'success',
            'data': error
        }

    def mag_align_start(self, *args):  # pylint: disable=unused-argument
        '''
        Start mag align action
        '''
        if not self.is_mag_align:
            self.is_mag_align = True

            thread = threading.Thread(
                target=self.thread_do_mag_align, args=())
            thread.start()

        return {
            'packetType': 'success'
        }

    @with_device_message
    def thread_do_mag_align(self):
        '''
        Do mag align
        '''
        try:
            command_line = helper.build_input_packet(
                'ma', self.properties, 'start')
            # self.communicator.write(command_line)
            # result = self.get_input_result('ma', timeout=3)
            result = yield self._message_center.build(command=command_line, timeout=3)

            time.sleep(1)
            has_result = False
            while not has_result:
                command_line = helper.build_input_packet(
                    'ma', self.properties, 'status')
                # self.communicator.write(command_line)
                # print('send ma status', command_line)
                result = yield self._message_center.build(command=command_line)
                # print(result['data'], self.is_mag_align)
                if not self.is_mag_align:
                    break
                # result = self.get_input_result('ma', timeout=1)
                # print('got ma status', result['data'])
                if result['data'] == b'\x00':
                    has_result = True
                else:
                    time.sleep(0.5)

            if not has_result:
                print('exit mag')
                return

            # print('ma status', result['data'])
            # print('mag status', result['data'])
            command_line = helper.build_input_packet(
                'ma', self.properties, 'stored')
            # self.communicator.write(command_line)
            # print('send ma stored', command_line)
            result = yield self._message_center.build(command=command_line)
            # print('ma stored result', result['data'])
            # result = self.get_input_result('ma', timeout=2)

            decoded_status = binascii.hexlify(result['data'])
            mag_value = self.decode_mag_align_output(decoded_status)
            self.is_mag_align = False

            # TODO: reset packet rate after operation successful
            self.add_output_packet('stream', 'mag_status', {
                'status': 'complete',
                'value': mag_value
            })
        except Exception:  # pylint: disable=broad-except
            self.is_mag_align = False
            self.add_output_packet('stream', 'mag_status', {
                'status': 'error'
            })

    @with_device_message
    def mag_align_abort(self, *args):  # pylint: disable=invalid-name
        '''
        Abort mag align action
        '''
        self.is_mag_align = False

        time.sleep(1)
        command_line = helper.build_input_packet(
            'ma', self.properties, 'abort')
        print('send mag abort', command_line)
        result = yield self._message_center.build(command=command_line)
        print('mag abort result', result['data'])
        # self.communicator.write(command_line)
        # result = self.get_input_result('ma', timeout=1)

        if result['error']:
            yield {
                'packetType': 'error',
                'data': {
                    'error': 1
                }
            }
        else:
            yield {
                'packetType': 'success'
            }

    @with_device_message
    def mag_align_save(self, *args):  # pylint: disable=invalid-name
        '''
        Save mag align resut
        '''
        command_line = helper.build_input_packet(
            'ma', self.properties, 'save')
        # self.communicator.write(command_line)
        # result = self.get_input_result('ma', timeout=1)
        result = yield self._message_center.build(command=command_line)

        if result['error']:
            yield {
                'packetType': 'error',
                'data': {
                    'error': 1
                }
            }

        yield {
            'packetType': 'success'
        }

    def decode_mag_align_output(self, value):
        '''
        decode mag align output
        '''
        hard_iron_x = dict()
        hard_iron_y = dict()
        soft_iron_ratio = dict()
        soft_iron_angle = dict()

        hard_iron_x['value'] = self.hard_iron_cal(value[16:20], 'axis')
        hard_iron_x['name'] = 'Hard Iron X'
        hard_iron_x['argument'] = 'hard_iron_x'

        hard_iron_y['value'] = self.hard_iron_cal(value[20:24], 'axis')
        hard_iron_y['name'] = 'Hard Iron Y'
        hard_iron_y['argument'] = 'hard_iron_y'

        soft_iron_ratio['value'] = self.hard_iron_cal(value[24:28], 'ratio')
        soft_iron_ratio['name'] = 'Soft Iron Ratio'
        soft_iron_ratio['argument'] = 'soft_iron_ratio'

        soft_iron_angle['value'] = self.hard_iron_cal(value[28:32], 'angle')
        soft_iron_angle['name'] = 'Soft Iron Angle'
        soft_iron_angle['argument'] = 'soft_iron_angle'

        output = [hard_iron_x, hard_iron_y, soft_iron_ratio, soft_iron_angle]

        return output

    def hard_iron_cal(self, value, data_type):
        '''
        convert hard iron value
        '''
        decoded_value = int(value, 16)
        # print (decodedValue)
        if data_type == 'axis':
            if decoded_value > 2 ** 15:
                new_decoded_value = (decoded_value - 2 ** 16)
                return new_decoded_value / float(2 ** 15) * 8
            else:
                return decoded_value / float(2 ** 15) * 8

        if data_type == 'ratio':
            return decoded_value / float(2 ** 16 - 1)

        if data_type == 'angle':
            if decoded_value > 2 ** 15:
                decoded_value = decoded_value - 2 ** 16
                pi_value = 2 ** 15 / math.pi
                return decoded_value / pi_value

            pi_value = 2 ** 15 / math.pi
            return decoded_value / pi_value

    def upgrade_framework(self, file, *args):  # pylint: disable=invalid-name
        '''
        upgrade framework
        '''
        # start a thread to do upgrade
        if not self.is_upgrading:
            self.is_upgrading = True
            self._message_center.pause()

            if self._logger is not None:
                self._logger.stop_user_log()

            thead = threading.Thread(
                target=self.thread_do_upgrade_framework, args=(file,))
            thead.start()
            print("Thread upgarde framework OpenIMU start at:[{0}].".format(
                datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')))
        return {
            'packetType': 'success'
        }

    def backup_calibration(self, params):
        '''
        start a thread to backup
        '''
        if not self.is_backup:
            self.is_backup = True
            self.ans_platform.set_access_token(params['token'])
            thread = threading.Thread(
                target=self.thread_do_backup, args=())
            thread.start()

        return {
            'packetType': 'success'
        }

    def restore_calibration(self, params):
        '''
        start a thread to restore
        '''
        # if not self.is_restore:
        #     self.is_restore = True
        #     self.ans_platform.set_access_token(params['token'])
        #     thread = threading.Thread(
        #         target=self.thread_do_restore, args=())
        #     thread.start()

        return {
            'packetType': 'success'
        }

    @with_device_message
    def thread_do_backup(self):
        '''
        Do Calibration Backup
        '''
        # get current odr
        packet_rate_param_index = 4
        command_line = helper.build_input_packet(
            'gP', properties=self.properties, param=packet_rate_param_index)
        packet_rate_result = yield self._message_center.build(command=command_line)

        if packet_rate_result['error']:
            self.is_backup = False
            self.add_output_packet('stream', 'backup_status', {
                'status': 'fail'
            })

            return

        # set quiet
        command_line = helper.build_input_packet(
            'uP', properties=self.properties, param=packet_rate_param_index, value=0)
        result = yield self._message_center.build(command=command_line)

        if packet_rate_result['error']:
            self.is_backup = False
            self.add_output_packet('stream', 'backup_status', {
                'status': 'fail'
            })

            return

        file_name = self.device_info['sn']+'.bin'  # todo: sn-yyyy-mm-dd-hhmmss
        start = 0x0
        read_size = 0x7E
        max_length = 4096
        file_write_size = 0
        file_result = bytearray()

        while file_write_size < max_length:
            actual_read_size = read_size
            plan_file_write_size = file_write_size + actual_read_size

            if plan_file_write_size >= max_length:
                actual_read_size = max_length - file_write_size

            command_line = helper.build_read_eeprom_input_packet(
                start, actual_read_size)
            result = yield self._message_center.build(command=command_line)

            if result['error']:
                self.is_backup = False
                self.add_output_packet('stream', 'backup_status', {
                    'status': 'fail'
                })
                break

            data = result['data']
            if plan_file_write_size >= max_length:
                file_result.extend(data[0:actual_read_size])
            else:
                file_result.extend(data)

            file_write_size += len(data)
            start += actual_read_size

        reserved_data = self._reserve_by_word(file_result)

        self._write_to_file(file_name, reserved_data)

        # restore odr
        command_line = helper.build_input_packet(
            'uP', properties=self.properties,
            param=packet_rate_param_index,
            value=packet_rate_result['data']['value'])
        yield self._message_center.build(command=command_line)

    def _reserve_by_word(self, data):
        start_index = 0x284
        reserved_data = bytearray()
        reserved_data.extend([00, 00])  # append 16 bit count of erases
        reserved_data.extend(data[0:start_index])
        need_reserve = data[start_index:]
        total_len = int((4095 - start_index)/2)
        for i in range(total_len):
            reserved_data.extend([need_reserve[i*2+1], need_reserve[i*2]])

        return reserved_data

    def _write_to_file(self, file_name, result):
        # save to local path, backup/{device_type}/{file_name}
        executor_path = resource.get_executor_path()
        backup_folder_path = os.path.join(
            executor_path, 'backup', 'openimu')
        file_path = os.path.join(backup_folder_path, file_name)
        if not os.path.isdir(backup_folder_path):
            os.makedirs(backup_folder_path)

        with open(file_path, 'wb') as file_stream:
            file_stream.write(result)

        stream = 'stream'
        backup_status = 'backup_status'
        status_complete = 'complete'
        status_fail = 'fail'

        try:
            config = get_config()
            account_name = config.AZURE_STORAGE_ACCOUNT
            container_name = config.AZURE_STORAGE_BACKUP_CONTAINER
            sas_token = self.ans_platform.get_sas_token()
            if sas_token == '':
                raise Exception('cannot get sas token')
            self.block_blob_service = BlockBlobService(account_name=account_name,
                                                       sas_token=sas_token,
                                                       protocol='http')
            self.block_blob_service.create_blob_from_path(container_name=container_name,
                                                          blob_name=file_name,
                                                          file_path=file_path)
        except Exception as ex:
            print('azure exception', ex)
            self.is_backup = False
            self.add_output_packet(stream, backup_status, {
                'status': status_fail
            })
            return

        # save to db
        serial_num = self.device_info['sn']
        save_result = self.ans_platform.save_backup_restult(
            serial_num, file_name, 'IMU')

        if save_result.__contains__('error'):
            self.is_backup = False
            self.add_output_packet(stream, backup_status, {
                'status': status_fail
            })
            return

        self.is_backup = False
        self.add_output_packet(stream, backup_status, {
            'status': status_complete,
            'date': save_result['data']['lastBackupTime']
        })

    @with_device_message
    def _unlock(self):
        command_line = helper.build_read_eeprom_input_packet(0x100, 2)
        result = yield self._message_center.build(command=command_line, timeout=2)
        if result['error']:
            self._restore_fail()
            return

        command_line = helper.build_unlock_eeprom_packet(result['data'])
        unlock_result = yield self._message_center.build(command=command_line, timeout=2)

        if unlock_result['error']:
            self._restore_fail()
            return
        print('unlock -- successfull', unlock_result['data'])

    @with_device_message
    def thread_do_restore(self):
        '''
        Do Calibration Restore
        '''
        # 1.download bin from azure
        file_name = self.device_info['sn']+'.bin'
        content_data = bytearray()
        executor_path = resource.get_executor_path()
        backup_folder_path = os.path.join(
            executor_path, 'backup', 'openimu')
        file_path = os.path.join(backup_folder_path, file_name)
        file = open(file_path, 'rb')
        content_data = file.read()
        file.close()
        # 2.save odr, then set quiet
        # get current odr
        packet_rate_param_index = 4
        command_line = helper.build_input_packet(
            'gP', properties=self.properties, param=packet_rate_param_index)
        packet_rate_result = yield self._message_center.build(command=command_line)

        if packet_rate_result['error']:
            self._restore_fail()
            return

        # set quiet
        command_line = helper.build_input_packet(
            'uP', properties=self.properties, param=packet_rate_param_index, value=0)
        result = yield self._message_center.build(command=command_line)

        if result['error']:
            self._restore_fail()
            return
        # 3.do restore
        # 3.1 unlock eeprom
        command_line = helper.build_read_eeprom_input_packet(0x100, 2)
        result = yield self._message_center.build(command=command_line, timeout=2)
        if result['error']:
            self._restore_fail()
            return

        command_line = helper.build_unlock_eeprom_packet(result['data'])
        unlock_result = yield self._message_center.build(command=command_line, timeout=2)

        if unlock_result['error']:
            self._restore_fail()
            return
        print('unlock -- successfull', unlock_result['data'])
        # 3.2 write eeprom
        skip_range = [0x200, 0x284]

        sn_string = self._build_sn_string(content_data[0x200, 0x204])
        model_string = self._build_model_string(content_data[0x204:0x284])

        can_write = self._clear_calibration_area()
        if not can_write:
            self._restore_fail()
            return

        can_write = self._write_calibration_from_data(content_data, skip_range)

        if not can_write:
            self._restore_fail()
            return

        can_write = self._write_sn_and_model_string(sn_string, model_string)

        if not can_write:
            self._restore_fail()
            return

        print('write eeporm -- successfull')

        self._lock()

        # 4.write operation result to db

        # 5.restore odr
        command_line = helper.build_input_packet(
            'uP', properties=self.properties,
            param=packet_rate_param_index,
            value=packet_rate_result['data']['value'])
        yield self._message_center.build(command=command_line)

        self.is_restore = False
        self.add_output_packet('stream', 'restore_status', {
            'status': 'success'
        })

    def _build_sn_string(self, data_range):
        data = []
        start = 0
        for _ in range(2):
            if start == 4:
                break
            data.extend([
                struct.unpack('B', data_range[start+1: start+2])[0],
                struct.unpack('B', data_range[start: start+1])[0]
            ])
            start += 2
        return data

    def _build_model_string(self, data_range):
        end = [0]
        data = []
        for item in data_range:
            if item == 0:
                break
            data.append(item)

        data.extend(end)
        return data

    @with_device_message
    def _clear_calibration_area(self):
        start = 0
        end = 4096
        block_size = 20
        write_offset = 0
        while write_offset < end:
            write_data = []
            plan_write_offset = write_offset + block_size * 2

            if plan_write_offset >= end:
                plan_write_offset = end
                block_size = int((plan_write_offset - write_offset)/2)

            for _ in range(plan_write_offset - write_offset):
                write_data.append(0xFF)
            command_line = helper.build_write_eeprom_input_packet(
                start, block_size, write_data)
            result = yield self._message_center.build(command=command_line, timeout=2)
            if result['error']:
                yield False

            write_offset = plan_write_offset
            start += block_size

        yield True

    @with_device_message
    def _write_sn_and_model_string(self, sn_string, model_string):
        command_line = helper.build_write_eeprom_input_packet(
            0x100, 2, sn_string)
        result = yield self._message_center.build(command=command_line, timeout=2)
        if result['error']:
            yield False

        command_line = helper.build_write_eeprom_input_packet(
            0x104, len(model_string), self._build_16bit_data_range(model_string))
        result = yield self._message_center.build(command=command_line, timeout=2)
        if result['error']:
            yield False

        yield True

    @with_device_message
    def _write_calibration_from_data(self, data, skip_range):
        end = 4096
        block_size = 20
        write_offset = 0

        while write_offset < end:
            plan_write_offset = write_offset + block_size * 2

            if plan_write_offset >= end:
                plan_write_offset = end
                block_size = int((plan_write_offset - write_offset)/2)

            # plan write range
            plan_write_range = [write_offset, plan_write_offset]
            # build a new range with skip range
            new_range = self._build_calibration_write_range(
                data, plan_write_range, skip_range)

            for write_range in new_range:
                command_line = helper.build_write_eeprom_input_packet(
                    int(write_range['start']/2),
                    int(write_range['length']/2),
                    write_range['data'])
                result = yield self._message_center.build(command=command_line, timeout=2)

                if result['error']:
                    yield False
        yield True

    def _build_calibration_write_range(self, content_data, plan_write_range, skip_range):
        # range struct {'start': start, 'data': data, 'length': length}
        new_range = []
        write_data = []
        if plan_write_range[1] >= skip_range[0] and plan_write_range[1] <= skip_range[1]:
            if plan_write_range[0] < skip_range[0]:
                write_data.extend(
                    content_data[plan_write_range[0]: skip_range[0]])
                new_range.append({'start': int(plan_write_range[0]/2),
                                  'length': int((skip_range[0]-plan_write_range[0])/2),
                                  'data': write_data})

        if plan_write_range[0] >= skip_range[0] and plan_write_range[0] <= skip_range[1]:
            if plan_write_range[1] > skip_range[1]:
                write_data.extend(
                    content_data[skip_range[1]: plan_write_range[1]])
                new_range.append({'start': int(skip_range[1]/2),
                                  'length': int((plan_write_range[1] - skip_range[1])/2),
                                  'data': write_data})

        if plan_write_range[0] < skip_range[0] and skip_range[1] < plan_write_range[1]:
            new_range.append({'start': int(plan_write_range[0]/2),
                              'length': int((skip_range[0] - plan_write_range[0])/2),
                              'data': content_data[plan_write_range[0]: skip_range[0]]})

            new_range.append({'start': int(skip_range[1]/2),
                              'length': int((plan_write_range[1] - skip_range[1])/2),
                              'data': content_data[skip_range[1]: plan_write_range[1]]})

        if plan_write_range[1] < skip_range[0] or plan_write_range[0] > skip_range[1]:
            write_data.extend(
                content_data[plan_write_range[0]: plan_write_range[1]])
            new_range.append({'start': int(plan_write_range[0]/2),
                              'length': int((plan_write_range[1] - plan_write_range[0])/2),
                              'data': write_data})

        return new_range

    def _build_16bit_data_range(self, data_range):
        data = []
        for item in data_range:
            data.extend(struct.pack('>H', item))
        return data

    def _build_reserve_data(self, data_range):
        data = []
        start = 0
        for _ in range(int(len(data_range)/2)):
            if start == len(data_range):
                break
            data.extend([
                struct.unpack('B', data_range[start+1: start+2])[0],
                struct.unpack('B', data_range[start: start+1])[0]
            ])
            start += 2
        return data

    @with_device_message
    def _lock(self):
        # lock eeprom
        command_line = helper.build_input_packet('LE')
        result = yield self._message_center.build(command=command_line)

        if result['error']:
            self._restore_fail()
            return
        print('lock eeporm -- successfull', result['data'])

        # software reset
        command_line = helper.build_input_packet('SR')
        yield self._message_center.build(command=command_line)

    def _restore_fail(self):
        self.is_restore = False
        self.add_output_packet('stream', 'restore_status', {
            'status': 'fail'
        })



import os
import time
import json
import datetime
import threading
import math
import serial
import serial.tools.list_ports
from .ntrip_client import NTRIPClient
from .sdk_upgrade import SDKUpgrade
from ...framework.utils import helper
from ...framework.utils import resource
from ...framework.context import APP_CONTEXT
from ..base.uart_base import OpenDeviceBase
from ..configs.openrtk_predefine import (
    APP_STR, get_app_names
)
from ..decorator import with_device_message
# import asyncio


class Provider(OpenDeviceBase):
    '''
    OpenRTK UART provider
    '''

    def __init__(self, communicator):
        super(Provider, self).__init__(communicator)
        self.type = 'RTK'
        self.server_update_rate = 100
        self.sky_data = []
        self.pS_data = []
        self.bootloader_baudrate = 115200
        self.app_config_folder = ''
        self.device_info = None
        self.app_info = None
        self.parameters = None
        self.setting_folder_path = None
        self.connection_file = None
        self.data_folder = None
        self.debug_serial_port = None
        self.rtcm_serial_port = None
        self.user_logf = None
        self.debug_logf = None
        self.rtcm_logf = None
        self.debug_c_f = None
        self.enable_data_log = False
        self.is_app_matched = False
        self.prepare_folders()

    def prepare_folders(self):
        '''
        Prepare folders for data storage and configuration
        '''
        executor_path = resource.get_executor_path()
        setting_folder_name = 'setting'
        config_file_name = 'openrtk.json'

        data_folder_path = os.path.join(executor_path, 'data')
        if not os.path.isdir(data_folder_path):
            os.makedirs(data_folder_path)
        self.data_folder = data_folder_path

        self.connection_file = os.path.join(
            executor_path, setting_folder_name, 'connection.json')

        # copy contents of app_config under executor path
        self.setting_folder_path = os.path.join(
            executor_path, setting_folder_name, 'openrtk')

        for app_name in get_app_names():
            app_name_path = os.path.join(self.setting_folder_path, app_name)
            app_name_config_path = os.path.join(
                app_name_path, config_file_name)
            if not os.path.isfile(app_name_config_path):
                if not os.path.isdir(app_name_path):
                    os.makedirs(app_name_path)
                app_config_content = resource.get_content_from_bundle(
                    setting_folder_name, os.path.join('openrtk', app_name, config_file_name))
                if app_config_content is None:
                    continue

                with open(app_name_config_path, "wb") as code:
                    code.write(app_config_content)

    def ping(self):
        '''
        Check if the connected device is OpenRTK
        '''
        # print('start to check if it is openrtk')
        device_info_text = self.internal_input_command('pG')
        app_info_text = self.internal_input_command('gV')

        APP_CONTEXT.get_logger().logger.debug('Checking if is OpenRTK device...')
        APP_CONTEXT.get_logger().logger.debug(
            'Device: {0}'.format(device_info_text))
        APP_CONTEXT.get_logger().logger.debug(
            'Firmware: {0}'.format(app_info_text))

        if device_info_text.find('OpenRTK') > -1:
            self.build_device_info(device_info_text)
            self.build_app_info(app_info_text)
            self.connected = True
            print('# Connected Information #')
            print('Device:', device_info_text)
            print('Firmware:', app_info_text)
            APP_CONTEXT.get_logger().logger.info(
                'Connected {0}, {1}'.format(device_info_text, app_info_text))
            return True
        return False

    def build_device_info(self, text):
        '''
        Build device info
        '''
        split_text = text.split(' ')
        self.device_info = {
            'name': split_text[0],
            'imu': split_text[1],
            'pn': split_text[2],
            'firmware_version': split_text[3],
            'sn': split_text[4]
        }

    def build_app_info(self, text):
        '''
        Build app info
        '''
        app_version = text

        split_text = app_version.split(' ')
        app_name = next(
            (item for item in APP_STR if item in split_text), None)

        if not app_name:
            app_name = 'INS'
            self.is_app_matched = False
        else:
            self.is_app_matched = True

        self.app_info = {
            'app_name': app_name,
            'version': text
        }

    def load_properties(self):
        # Load config from user working path
        local_config_file_path = os.path.join(os.getcwd(), 'openrtk.json')
        if os.path.isfile(local_config_file_path):
            with open(local_config_file_path) as json_data:
                self.properties = json.load(json_data)
                return

        # Load the openimu.json based on its app
        app_name = self.app_info['app_name']
        app_file_path = os.path.join(
            self.setting_folder_path, app_name, 'openrtk.json')

        with open(app_file_path) as json_data:
            self.properties = json.load(json_data)

    def ntrip_client_thread(self):
        self.ntripClient = NTRIPClient(self.properties, self.communicator)
        self.ntripClient.run()

    def build_connected_serial_port_info(self):
        if not os.path.isfile(self.connection_file):
            return None, None

        with open(self.connection_file) as json_data:
            connection = json.load(json_data)

        user_port = connection['port']
        user_port_num = ''
        port_name = ''
        for i in range(len(user_port)-1, -1, -1):
            if (user_port[i] >= '0' and user_port[i] <= '9'):
                user_port_num = user_port[i] + user_port_num
            else:
                port_name = user_port[:i+1]
                break
        return user_port_num, port_name

    def after_setup(self):
        set_user_para = self.cli_options and self.cli_options.set_user_para
        self.ntrip_client_enable = self.cli_options and self.cli_options.ntrip_client
        # with_raw_log = self.cli_options and self.cli_options.with_raw_log

        if set_user_para:
            result = self.set_params(
                self.properties["initial"]["userParameters"])
            ##print('set user para {0}'.format(result))
            if (result['packetType'] == 'success'):
                self.save_config()

        if self.ntrip_client_enable:
            t = threading.Thread(target=self.ntrip_client_thread)
            t.start()

        # if with_raw_log:
        connection = None
        debug_port = ''
        rtcm_port = ''
        try:
            if (self.properties["initial"]["useDefaultUart"]):
                user_port_num, port_name = self.build_connected_serial_port_info()
                if not user_port_num or not port_name:
                    return False
                #print('user_port {0} {1}'.format(user_port_num, port_name))
                debug_port = port_name + str(int(user_port_num) + 2)
                rtcm_port = port_name + str(int(user_port_num) + 1)
            else:
                for x in self.properties["initial"]["uart"]:
                    if x['enable'] == 1:
                        if x['name'] == 'DEBUG':
                            debug_port = x["value"]
                        elif x['name'] == 'GNSS':
                            rtcm_port = x["value"]

            if self.data_folder is not None:
                dir_time = time.strftime("%Y%m%d_%H%M%S", time.localtime())
                file_time = time.strftime(
                    "%Y_%m_%d_%H_%M_%S", time.localtime())
                file_name = self.data_folder + '/' + 'openrtk_log_' + dir_time
                os.mkdir(file_name)
                self.user_logf = open(
                    file_name + '/' + 'user_' + file_time + '.bin', "wb")

            if rtcm_port != '':
                print('OpenRTK log GNSS UART {0}'.format(rtcm_port))
                self.rtcm_serial_port = serial.Serial(
                    rtcm_port, '460800', timeout=0.005)
                if self.rtcm_serial_port.isOpen():
                    self.rtcm_logf = open(
                        file_name + '/' + 'rtcm_rover_' + file_time + '.bin', "wb")
                    t = threading.Thread(
                        target=self.thread_rtcm_port_receiver, args=(file_name,))
                    t.start()

            if debug_port != '':
                print('OpenRTK log DEBUG UART {0}'.format(debug_port))
                self.debug_serial_port = serial.Serial(
                    debug_port, '460800', timeout=0.005)
                if self.debug_serial_port.isOpen():
                    if self.app_info['app_name'] == 'RAWDATA':
                        self.debug_logf = open(
                            file_name + '/' + 'rtcm_base_' + file_time + '.bin', "wb")
                    elif self.app_info['app_name'] == 'RTK':
                        self.debug_logf = open(
                            file_name + '/' + 'rtcm_base_' + file_time + '.bin', "wb")
                    else:
                        self.debug_logf = open(
                            file_name + '/' + 'debug_' + file_time + '.bin', "wb")
                    t = threading.Thread(
                        target=self.thread_debug_port_receiver, args=(file_name,))
                    t.start()

        except Exception as e:
            if self.debug_serial_port is not None:
                if self.debug_serial_port.isOpen():
                    self.debug_serial_port.close()
            if self.rtcm_serial_port is not None:
                if self.rtcm_serial_port.isOpen():
                    self.rtcm_serial_port.close()
            self.debug_serial_port = None
            self.rtcm_serial_port = None
            print(e)
            return False

    def on_read_raw(self, data):
        if self.user_logf is not None:
            self.user_logf.write(data)

    def thread_debug_port_receiver(self, *args, **kwargs):
        if self.debug_logf is None:
            return

        is_get_configuration = 0
        file_name = args[0]
        self.debug_c_f = open(file_name + '/' + 'configuration.txt', "w")

        while True:
            if is_get_configuration:
                break
            cmd_configuration = 'get configuration\r\n'
            self.debug_serial_port.write(cmd_configuration.encode())
            try_times = 20
            for i in range(try_times):
                data_buffer = self.debug_serial_port.read(500)
                if len(data_buffer):
                    try:
                        #print('len = {0}'.format(len(data_buffer)))
                        str_data = bytes.decode(data_buffer)
                        # print('{0}'.format(str_data))
                        json_data = json.loads(str_data)
                        for key in json_data.keys():
                            if key == 'openrtk configuration':
                                print('{0}'.format(json_data))
                                if self.debug_c_f:
                                    self.debug_c_f.write(str_data)
                                    self.debug_c_f.close()
                                is_get_configuration = 1
                        if is_get_configuration:
                            break
                    except Exception as e:
                        #print('DEBUG PORT Thread:json error:', e)
                        # the json will not be completed
                        pass

        cmd_log = 'log debug on\r\n'
        self.debug_serial_port.write(cmd_log.encode())

        # log data
        while True:
            try:
                data = bytearray(self.debug_serial_port.read_all())
            except Exception as e:
                print('DEBUG PORT Thread:receiver error:', e)
                return  # exit thread receiver
            if len(data):
                self.debug_logf.write(data)
            else:
                time.sleep(0.001)

    def thread_rtcm_port_receiver(self, *args, **kwargs):
        if self.rtcm_logf is None:
            return
        while True:
            try:
                data = bytearray(self.rtcm_serial_port.read_all())
            except Exception as e:
                print('RTCM PORT Thread:receiver error:', e)
                return  # exit thread receiver
            if len(data):
                self.rtcm_logf.write(data)
            else:
                time.sleep(0.001)

    def on_receive_output_packet(self, packet_type, data, error=None):
        '''
        Listener for getting output packet
        '''
        # $GPGGA,080319.00,3130.4858508,N,12024.0998832,E,4,25,0.5,12.459,M,0.000,M,2.0,*46
        if packet_type == 'gN':
            if self.ntrip_client_enable:
                # $GPGGA
                gpgga = '$GPGGA'
                # time
                timeOfWeek = float(data['GPS_TimeofWeek'])
                dsec = int(timeOfWeek)
                msec = timeOfWeek - dsec
                sec = dsec % 86400
                hour = int(sec / 3600)
                minute = int(sec % 3600 / 60)
                second = sec % 60
                gga_time = format(hour*10000 + minute*100 +
                                  second + msec - 18, '09.2f')
                gpgga = gpgga + ',' + gga_time
                # latitude
                latitude = float(data['latitude']) / 1e+07
                if latitude >= 0:
                    latflag = 'N'
                else:
                    latflag = 'S'
                    latitude = math.fabs(latitude)
                lat_d = int(latitude)
                lat_m = (latitude-lat_d) * 60
                lat_dm = format(lat_d*100 + lat_m, '012.7f')
                gpgga = gpgga + ',' + lat_dm + ',' + latflag
                # longitude
                longitude = float(data['longitude']) / 1e+07
                if longitude >= 0:
                    lonflag = 'E'
                else:
                    lonflag = 'W'
                    longitude = math.fabs(longitude)
                lon_d = int(longitude)
                lon_m = (longitude-lon_d) * 60
                lon_dm = format(lon_d*100 + lon_m, '013.7f')
                gpgga = gpgga + ',' + lon_dm + ',' + lonflag
                # positionMode
                gpgga = gpgga + ',' + str(data['positionMode'])
                # svs
                gpgga = gpgga + ',' + str(data['numberOfSVs'])
                # hop
                gpgga = gpgga + ',' + format(float(data['hdop']), '03.1f')
                # height
                gpgga = gpgga + ',' + \
                    format(float(data['height']), '06.3f') + ',M'
                #
                gpgga = gpgga + ',0.000,M'
                # diffage
                gpgga = gpgga + ',' + \
                    format(float(data['diffage']), '03.1f') + ','
                # ckm
                checksum = 0
                for i in range(1, len(gpgga)):
                    checksum = checksum ^ ord(gpgga[i])
                gpgga = gpgga + '*' + str(checksum) + '\r\n'
                print(gpgga)
                self.ntripClient.send(gpgga)
                return

        elif packet_type == 'pS':
            if data['latitude'] != 0.0 and data['longitude'] != 0.0:
                if self.pS_data:
                    if self.pS_data['GPS_Week'] == data['GPS_Week']:
                        if data['GPS_TimeofWeek'] - self.pS_data['GPS_TimeofWeek'] >= 0.2:
                            self.add_output_packet('stream', 'pos', data)
                            self.pS_data = data
                    else:
                        self.add_output_packet('stream', 'pos', data)
                        self.pS_data = data
                else:
                    self.add_output_packet('stream', 'pos', data)
                    self.pS_data = data

        elif packet_type == 'sK':
            if self.sky_data:
                if self.sky_data[0]['timeOfWeek'] == data[0]['timeOfWeek']:
                    self.sky_data.extend(data)
                else:
                    self.add_output_packet('stream', 'skyview', self.sky_data)
                    self.add_output_packet('stream', 'snr', self.sky_data)
                    self.sky_data = []
                    self.sky_data.extend(data)
            else:
                self.sky_data.extend(data)

        else:
            output_packet_config = next(
                (x for x in self.properties['userMessages']['outputPackets']
                 if x['name'] == packet_type), None)
            if output_packet_config and output_packet_config.__contains__('from') \
                    and output_packet_config['from'] == 'imu':
                self.add_output_packet('stream', 'imu', data)

    # command list
    def server_status(self, *args):  # pylint: disable=invalid-name
        '''
        Get server connection status
        '''
        return {
            'packetType': 'ping',
            'data': {'status': '1'}
        }

    def get_device_info(self, *args):  # pylint: disable=invalid-name
        '''
        Get device information
        '''
        return {
            'packetType': 'deviceInfo',
            'data':  [
                {'name': 'Product Name', 'value': self.device_info['name']},
                {'name': 'IMU', 'value': self.device_info['imu']},
                {'name': 'PN', 'value': self.device_info['pn']},
                {'name': 'Firmware Version',
                 'value': self.device_info['firmware_version']},
                {'name': 'SN', 'value': self.device_info['sn']},
                {'name': 'App Version', 'value': self.app_info['version']}
            ]
        }

    def get_log_info(self):
        '''
        Build information for log
        '''
        return {
            "type": self.type,
            "model": self.device_info['name'],
            "logInfo": {
                "pn": self.device_info['pn'],
                "sn": self.device_info['sn'],
                "rtkProperties": json.dumps(self.properties)
            }
        }

    def get_conf(self, *args):  # pylint: disable=unused-argument
        '''
        Get json configuration
        '''
        return {
            'packetType': 'conf',
            'data': {
                'outputs': self.properties['userMessages']['outputPackets'],
                'inputParams': self.properties['userConfiguration']
            }
        }

    @with_device_message
    def get_params(self, *args):  # pylint: disable=unused-argument
        '''
        Get all parameters
        '''
        has_error = False
        parameter_values = []

        if self.app_info['app_name'] == 'INS':
            conf_parameters = self.properties['userConfiguration']
            conf_parameters_len = len(conf_parameters)-1
            step = 10

            for i in range(2, conf_parameters_len, step):
                start_byte = i
                end_byte = i+step-1 if i+step < conf_parameters_len else conf_parameters_len

                command_line = helper.build_packet(
                    'gB', [start_byte, end_byte])
                result = yield self._message_center.build(command=command_line, timeout=2)
                if result['error']:
                    has_error = True
                    break

                parameter_values.extend(result['data'])
        else:
            command_line = helper.build_input_packet('gA')
            result = yield self._message_center.build(command=command_line, timeout=3)
            if result['error']:
                has_error = True

            parameter_values = result['data']

        if not has_error:
            self.parameters = parameter_values
            yield {
                'packetType': 'inputParams',
                'data': parameter_values
            }

        yield {
            'packetType': 'error',
            'data': 'No Response'
        }

    @with_device_message
    def get_param(self, params, *args):  # pylint: disable=unused-argument
        '''
        Update paramter value
        '''
        command_line = helper.build_input_packet(
            'gP', properties=self.properties, param=params['paramId'])
        # self.communicator.write(command_line)
        # result = self.get_input_result('gP', timeout=1)
        result = yield self._message_center.build(command=command_line)

        if result['data']:
            self.parameters = result['data']
            yield {
                'packetType': 'inputParam',
                'data': result['data']
            }

        yield {
            'packetType': 'error',
            'data': 'No Response'
        }

    @with_device_message
    def set_params(self, params, *args):  # pylint: disable=unused-argument
        '''
        Update paramters value
        '''
        for parameter in params:
            # result = self.set_param(parameter)
            command_line = helper.build_input_packet(
                'uP', properties=self.properties,
                param=parameter['paramId'],
                value=parameter['value'])

            result = yield self._message_center.build(command=command_line)

            packet_type = result['packet_type']
            data = result['data']

            if packet_type == 'error':
                yield {
                    'packetType': 'error',
                    'data': {
                        'error': data
                    }
                }
            if data > 0:
                yield {
                    'packetType': 'error',
                    'data': {
                        'error': data
                    }
                }

        yield {
            'packetType': 'success',
            'data': {
                'error': 0
            }
        }

    @with_device_message
    def set_param(self, params, *args):  # pylint: disable=unused-argument
        '''
        Update paramter value
        '''
        command_line = helper.build_input_packet(
            'uP', properties=self.properties, param=params['paramId'], value=params['value'])
        # self.communicator.write(command_line)
        # result = self.get_input_result('uP', timeout=1)
        result = yield self._message_center.build(command=command_line)

        error = result['error']
        data = result['data']
        if error:
            yield {
                'packetType': 'error',
                'data': {
                    'error': data
                }
            }

        yield {
            'packetType': 'success',
            'data': {
                'error': data
            }
        }

    @with_device_message
    def save_config(self, *args):  # pylint: disable=unused-argument
        '''
        Save configuration
        '''
        command_line = helper.build_input_packet('sC')
        # self.communicator.write(command_line)
        # result = self.get_input_result('sC', timeout=2)
        result = yield self._message_center.build(command=command_line, timeout=2)

        data = result['data']
        error = result['error']
        if data:
            yield {
                'packetType': 'success',
                'data': error
            }

        yield {
            'packetType': 'success',
            'data': error
        }

    @with_device_message
    def reset_params(self, params, *args):  # pylint: disable=unused-argument
        '''
        Reset params to default
        '''
        command_line = helper.build_input_packet('rD')
        result = yield self._message_center.build(command=command_line, timeout=2)

        error = result['error']
        data = result['data']
        if error:
            yield {
                'packetType': 'error',
                'data': {
                    'error': error
                }
            }

        yield {
            'packetType': 'success',
            'data': data
        }

    def upgrade_framework(self, params, *args):  # pylint: disable=unused-argument
        '''
        Upgrade framework
        '''
        app_name = ''
        file = ''
        if isinstance(params, str):
            file = params

        if isinstance(params, dict):
            app_name = params['name']
            file = params['file']

        # app_name='GNSS_RTK_SDK'
        # start a thread to do upgrade
        if not self.is_upgrading:
            self.is_upgrading = True
            self._message_center.pause()

            if self._logger is not None:
                self._logger.stop_user_log()

            if app_name.__contains__('SDK'):
                thread = threading.Thread(
                    target=self.thread_do_sdk_framework, args=(file,))
                thread.start()
                print("Thread upgarde sdk of OpenRTK start at:[{0}].".format(
                    datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')))
            else:
                thread = threading.Thread(
                    target=self.thread_do_upgrade_framework, args=(file,))
                thread.start()
                print("Thread upgarde framework of OpenRTK start at:[{0}].".format(
                    datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')))
        return {
            'packetType': 'success'
        }

    def thread_do_sdk_framework(self, file):
        '''
        Do upgrade firmware
        '''
        try:
            file_content = self._do_download_firmware(file)
            user_port_num, port_name = self.build_connected_serial_port_info()
            sdk_port = port_name + str(int(user_port_num) + 3)
            sdk_upgrade = SDKUpgrade(sdk_port, file_content)
            sdk_upgrade.on('error', self.on_sdk_upgarde_error)
            # sdk_upgrade.on('start', self.on_sdk_upgrade_start)
            sdk_upgrade.on('progress', self.on_sdk_upgarde_progress)
            sdk_upgrade.on('finish', self.on_sdk_upgarde_finish)
            sdk_upgrade.start()
        except Exception as ex:  # pylint:disable=broad-except
            self.on_upgarde_failed('Upgrade Failed')
            APP_CONTEXT.get_logger().logger.error(
                'Upgrade Error: {0}'.format(ex))

    def on_sdk_upgarde_error(self, message):
        print('upgrade error', message)
        self.is_upgrading = False
        self._message_center.resume()
        self.add_output_packet(
            'stream', 'upgrade_complete', {'success': False, 'message': message})

    def on_sdk_upgrade_start(self, addr, fs_len):
        pass

    def on_sdk_upgarde_progress(self, addr, fs_len):
        self.add_output_packet('stream', 'upgrade_progress', {
            'addr': addr,
            'fs_len': fs_len
        })

    def on_sdk_upgarde_finish(self, message):
        self.is_upgrading = False
        self._message_center.resume()
        self.add_output_packet(
            'stream', 'upgrade_complete', {'success': True})



import struct

IIR_50HZ_LPF = 50
IIR_40HZ_LPF = 40
IIR_25HZ_LPF = 25
IIR_20HZ_LPF = 20
IIR_10HZ_LPF = 10
IIR_05HZ_LPF = 5
IIR_02HZ_LPF = 2
UNFILTERED = 0


def bytes_to_byte_instr(b, n=None):
    s = ''.join(f'{x:08b}' for x in b)
    return s if n is None else s[:n + n // 8 + (0 if n % 8 else -1)]


def decode_lpf(payload):
    # print('decode lpf', payload)
    pack_item = struct.pack('2B', *payload)
    counts = struct.unpack('>H', pack_item)[0]
    if counts > 18749:
        return IIR_02HZ_LPF
    elif (counts <= 18749) and (counts > 8034):
        return IIR_05HZ_LPF
    elif (counts <= 8034) and (counts > 4017):
        return IIR_10HZ_LPF
    elif (counts <= 4017) and (counts > 2410):
        return IIR_20HZ_LPF
    elif (counts <= 2410) and (counts > 1740):
        return IIR_25HZ_LPF
    elif (counts <= 1740) and (counts > 1204):
        return IIR_40HZ_LPF
    elif (counts <= 1204) and (counts > 0):
        return IIR_50HZ_LPF
    else:
        return UNFILTERED


def encode_lpf(value):
    # print('encode lpf', value)
    counts = 0
    if value == IIR_02HZ_LPF:
        counts = 26785
    elif value == IIR_05HZ_LPF:
        counts = 10713
    elif value == IIR_10HZ_LPF:
        counts = 5356
    elif value == IIR_20HZ_LPF:
        counts = 2678
    elif value == IIR_25HZ_LPF:
        counts = 2142
    elif value == IIR_40HZ_LPF:
        counts = 1338
    elif value == UNFILTERED:
        counts = 0
    else:
        counts = 1070

    return struct.pack('>h', counts)


def decode_value(data_type, data):
    if data_type == 'uint64':
        try:
            pack_item = struct.pack('8B', *data)
        except:  # pylint: disable=bare-except
            return False
        return struct.unpack('>Q', pack_item)[0]
    elif data_type == 'int64':
        try:
            pack_item = struct.pack('8B', *data)
        except:  # pylint: disable=bare-except
            return False
        return struct.unpack('>q', pack_item)[0]
    elif data_type == 'double':
        try:
            pack_item = struct.pack('8B', *data)
        except:  # pylint: disable=bare-except
            return False
        return struct.unpack('d', pack_item)[0]
    elif data_type == 'uint32':
        try:
            pack_item = struct.pack('4B', *data)
        except:  # pylint: disable=bare-except
            return False
        return struct.unpack('>I', pack_item)[0]
    elif data_type == 'int32':
        try:
            pack_item = struct.pack('4B', *data)
        except:  # pylint: disable=bare-except
            return False
        return struct.unpack('>i', pack_item)[0]
    elif data_type == 'float':
        try:
            pack_item = struct.pack('4B', *data)
        except:  # pylint: disable=bare-except
            return False
        return struct.unpack('>f', pack_item)[0]
    elif data_type == 'uint16':
        try:
            pack_item = struct.pack('2B', *data)
        except:  # pylint: disable=bare-except
            return False
        return struct.unpack('>H', pack_item)[0]
    elif data_type == 'int16':
        try:
            pack_item = struct.pack('2B', *data)
        except:  # pylint: disable=bare-except
            return False
        return struct.unpack('>h', pack_item)[0]
    elif data_type == 'uint8':
        try:
            pack_item = struct.pack('1B', *data)
        except:  # pylint: disable=bare-except
            return False
        return struct.unpack('>B', pack_item)[0]
    elif data_type == 'int8':
        try:
            pack_item = struct.pack('1B', *data)
        except:  # pylint: disable=bare-except
            return False
        return struct.unpack('>b', pack_item)[0]
    elif 'char' in data_type:
        try:
            trim_data = [elem for elem in data if elem != 0]
            max_len = int(data_type.replace('char', ''))
            trim_len = len(trim_data)
            data_len = trim_len if trim_len < max_len else max_len
            pack_item = struct.pack('{0}B'.format(data_len), *trim_data)
            return pack_item.decode()
        except:  # pylint: disable=bare-except
            return False
    elif data_type == 'string':
        try:
            fmt_str = '{0}B'.format(len(data))
            pack_item = struct.pack(fmt_str, *data)
            return pack_item
        except:  # pylint: disable=bare-except
            return False
    elif data_type == 'ip4':
        try:
            ip_1 = str(data[0])
            ip_2 = str(data[1])
            ip_3 = str(data[2])
            ip_4 = str(data[3])
            return ip_1+'.'+ip_2+'.'+ip_3+'.'+ip_4
        except:  # pylint: disable=bare-except
            return False
    elif data_type == 'ip6':
        try:
            ip_1 = str(data[0])
            ip_2 = str(data[1])
            ip_3 = str(data[2])
            ip_4 = str(data[3])
            ip_5 = str(data[4])
            ip_6 = str(data[5])
            return ip_1+'.'+ip_2+'.'+ip_3+'.'+ip_4+'.'+ip_5+'.'+ip_6
        except:  # pylint: disable=bare-except
            return False
    elif data_type == 'orientation':
        byte_str = bytes_to_byte_instr(data)
        ret_data = ''
        x_sign = byte_str[-1]
        ret_data += '+' if x_sign == '0' else '-'
        x_asix = int(byte_str[-3:-1], 2)
        ret_data += ['X', 'Y', 'Z'][x_asix]

        y_sign = byte_str[-4]
        ret_data += '+' if y_sign == '0' else '-'
        y_asix = int(byte_str[-6:-4], 2)
        ret_data += ['Y', 'Z', 'X'][y_asix]

        z_sign = byte_str[-7]
        ret_data += '+' if z_sign == '0' else '-'
        z_asix = int(byte_str[-9:-7], 2)
        ret_data += ['Z', 'X', 'Y'][z_asix]

        return ret_data  # '+X+Y+Z'
    elif data_type == 'lpf':
        return decode_lpf(data)
    else:
        return False


def encode_value(data_type, data):
    if 'char' in data_type:
        actual_len = len(data)
        max_len = int(data_type.replace('char', ''))
        data_len = actual_len if actual_len < max_len else max_len
        bytes_data = bytearray(data, 'utf-8')
        return struct.pack('{0}B'.format(data_len), *bytes_data)

    elif data_type == 'uint16':
        return struct.pack('>h', data)

    elif data_type == 'orientation':
        if len(data) == 6:
            bit_str = ''
            x_sign = data[0]
            bit_str += '0' if x_sign == '+' else '1'
            x_asix = data[1]
            bit_str = {'X': '00', 'Y': '01', 'Z': '10'}.get(x_asix) + bit_str

            y_sign = data[2]
            bit_str = ('0' if y_sign == '+' else '1') + bit_str
            y_asix = data[3]
            bit_str = {'Y': '00', 'Z': '01', 'X': '10'}.get(y_asix) + bit_str

            z_sign = data[4]
            bit_str = ('0' if z_sign == '+' else '1') + bit_str
            z_asix = data[5]
            bit_str = {'Z': '00', 'X': '01', 'Y': '10'}.get(z_asix) + bit_str

            value = int(bit_str, 2)
            return struct.pack('>h', value)
        return False
    elif data_type == 'lpf':
        return encode_lpf(data)
    else:
        return False



from socket import *
import concurrent.futures as futures
import time
import base64

class NTRIPClient:
    def __init__(self, properties, communicator):
        self.communicator = communicator
        self.isConnected = 0
        for x in properties["initial"]["ntrip"]:
            if x['name'] == 'ip':
                self.ip = x["value"]
            elif x['name'] == 'port':
                self.port = x["value"]
            elif x['name'] == 'mountPoint':
                self.mountPoint = x["value"]
            elif x['name'] == 'username':
                self.username = x["value"]
            elif x['name'] == 'password':
                self.password = x["value"]
        
    def run(self):
        while True:
            while True:
                self.isConnected = self.doConnect()
                if self.isConnected == 0:
                    time.sleep(3)
                else:
                    break
            recvData = self.recvResponse()
            # print(recvData)
            if recvData != None and recvData.find('ICY 200 OK') != -1:
                print('NTRIP:[request] ok')
                self.recv()
            else:
                self.tcpClientSocket.close()
                
    def doConnect(self):
        self.isConnected = 0
        self.tcpClientSocket = socket(AF_INET, SOCK_STREAM)
        try:
            print('NTRIP:[connect] {0}:{1} start...'.format(self.ip, self.port))
            self.tcpClientSocket.connect((self.ip, self.port))
            print('NTRIP:[connect] ok')
            self.isConnected = 1
        except Exception as e:
            print('NTRIP:[connect] {0}'.format(e))
        if self.isConnected == 1:
            # send ntrip request
            ntripRequestStr = 'GET /' + self.mountPoint + ' HTTP/1.1\r\n'
            ntripRequestStr = ntripRequestStr + 'User-Agent: NTRIP Aceinna/0.1\r\n'
            ntripRequestStr = ntripRequestStr + 'Authorization: Basic '
            apikey = self.username + ':' + self.password
            apikeyBytes = apikey.encode("utf-8")
            ntripRequestStr = ntripRequestStr + base64.b64encode(apikeyBytes).decode("utf-8")
            ntripRequestStr = ntripRequestStr + '\r\n\r\n'
            # print(ntripRequestStr)
            self.send(ntripRequestStr)
        return self.isConnected

    def send(self, data):
        if self.isConnected:
            try:
                if isinstance(data, str):
                    self.tcpClientSocket.send(data.encode('utf-8'))
                else:
                    self.tcpClientSocket.send(bytes(data))
            except Exception as e:
                print('NTRIP:[send] error occur {0}'.format(e))
        else:
            return

    def recv(self):
        self.tcpClientSocket.settimeout(None)
        while True:
            try:
                data = self.tcpClientSocket.recv(1024)
                if data:
                    self.communicator.write(data)
            except Exception as e:
                print('NTRIP:[recv] error occur {0}'.format(e))
                self.tcpClientSocket.close()
                return

    def recvResponse(self):
        self.tcpClientSocket.settimeout(3)
        while True:
            try:
                data = self.tcpClientSocket.recv(1024)
                if data:
                    return data.decode('utf-8')
                else:
                    return
            except Exception as e:
                print('NTRIP:[recv] error occur {0}'.format(e))
                return
                


import os
import sys
import pkgutil
from ... import PACKAGE_NAME


def is_in_bundle():
    return hasattr(sys, 'frozen') and getattr(sys, 'frozen') and hasattr(sys, '_MEIPASS')


def is_dev_mode():
    return hasattr(sys, '__dev__') and getattr(sys, '__dev__')


def get_executor_path():
    if is_in_bundle():
        path = os.path.abspath(os.path.dirname(sys.executable))
    else:
        if is_dev_mode():  # if start from main.py
            path = os.getcwd()
        else:
            path = os.path.join(os.path.expanduser('~'), PACKAGE_NAME)
            if not os.path.isdir(path):
                os.makedirs(path)
    return path


def get_content_from_bundle(package, path):
    module_name = 'aceinna'
    if is_in_bundle():
        content = pkgutil.get_data(package, path)
    else:
        content = pkgutil.get_data(module_name, os.path.join(package, path))

    return content



from .openimu.uart_provider import Provider as OpenIMUProvider
from .openrtk.uart_provider import Provider as OpenRTKProvider
from .dmu.uart_provider import Provider as DMUProvider


class DeviceManager:
    '''
    Manage devices
    '''
    device_list = []

    @staticmethod
    def ping(communicator, *args):
        '''
        Find the matched device
        '''
        filter_device_type = args[0]

        if communicator.type == 'uart':
            if len(DeviceManager.device_list) == 0:
                DeviceManager.device_list.append(OpenIMUProvider(communicator))
                DeviceManager.device_list.append(OpenRTKProvider(communicator))
                DeviceManager.device_list.append(DMUProvider(communicator))

        for device in DeviceManager.device_list:
            if filter_device_type and device.type != filter_device_type:
                continue

            can_ping = device.ping()
            if can_ping:
                return device
        return None



from ..models.args import DetectorArgs
from ..framework.communicator import CommunicatorFactory

class Detector(object):
    '''
    Device Detector
    '''
    def __init__(self, **kwargs):
        self.communication = 'uart'
        self.communicator = None
        self._build_options(**kwargs)

    def find(self, callback):
        '''find if there is a connected device'''
        print('start to find device')
        if self.communicator is None:
            self.communicator = CommunicatorFactory.create(
                self.communication, self.options)

        self.communicator.find_device(callback)

    def _build_options(self, **kwargs):
        self.options = DetectorArgs(**kwargs)



"""
Application Loader
"""
from .cli import CommandLine
from .web import Webserver
#from ..framework.context import APP_CONTEXT


class Loader:
    '''Bootstrap Factory
    '''
    @staticmethod
    def create(platform, options):
        '''Initial bootstrap instance
        '''
        if platform == 'web':
            active_app = Webserver(**options)

        if platform == 'cli':
            active_app = CommandLine(**options)

        if active_app is None:
            raise Exception('no matched bootstrap')

        #APP_CONTEXT.set_app(active_app)

        return active_app



import os
import time
import json
import struct
from ..base.uart_base import OpenDeviceBase
from ..decorator import with_device_message
from ...framework.utils import (helper, resource)
from ...framework.ans_platform_api import AnsPlatformAPI
from . import dmu_helper
from .configuration_field import CONFIGURATION_FIELD_DEFINES_SINGLETON
from .eeprom_field import (EEPROM_ADDRESS_DEFINES,
                           EEPROM_FIELD_DEFINES_SINGLETON)

ID = [0x49, 0x44]
VR = [0x56, 0x52]
S0 = [0x53, 0x30]


class Provider(OpenDeviceBase):
    '''
    DMU UART provider
    '''

    def __init__(self, communicator):
        super(Provider, self).__init__(communicator)
        self.type = 'DMU'
        self.server_update_rate = 50
        self.is_logging = False
        self.is_mag_align = False
        self.bootloader_baudrate = 57600
        self.device_info = None
        self.app_info = None
        self.app_config_folder = ''
        self.parameters = None
        self.enable_data_log = True
        self.is_backup = False
        self.is_restore = False
        self.is_app_matched = False
        self.is_conf_loaded = False
        self.ans_platform = AnsPlatformAPI()
        self.prepare_folders()

    def prepare_folders(self):
        '''
        Prepare folder
        '''
        '''
        Prepare folders for data storage and configuration
        '''
        executor_path = resource.get_executor_path()
        setting_folder_name = 'setting'
        config_file_name = 'dmu.json'
        data_folder_path = os.path.join(executor_path, 'data')
        if not os.path.isdir(data_folder_path):
            os.makedirs(data_folder_path)

        self.setting_folder_path = os.path.join(
            executor_path, setting_folder_name, 'dmu')

        config_file_path = os.path.join(
            self.setting_folder_path, config_file_name)

        if not os.path.isfile(config_file_path):
            if not os.path.isdir(self.setting_folder_path):
                os.makedirs(self.setting_folder_path)

            app_config_content = resource.get_content_from_bundle(
                setting_folder_name, os.path.join('dmu', config_file_name))

            with open(config_file_path, "wb") as code:
                code.write(app_config_content)

    def ping(self):
        '''
        ping
        '''
        command = 'PK'
        command_line = dmu_helper.build_packet(command)
        self.communicator.write(command_line)
        time.sleep(0.1)
        data_buffer = self.read_untils_have_data(command)

        if data_buffer == []:
            # consider as dmu
            has_device_info = self._build_device_info()
            has_app_info = self._build_app_info()
            if not has_app_info or not has_device_info:
                return False

            #print('The device work as DMU')
            print('# Connected Information #')
            print('Device SN: {0}'.format(self.device_info['sn']))
            print('Device Model: {0} {1}'.format(
                self.device_info['name'], self.device_info['pn']))
            print('Firmware Version:', self.app_info['version'])
            self.connected = True
            self.is_conf_loaded = False
            return True
        return False

    def _build_device_info(self):
        '''
        Build device info
        '''
        command_line = dmu_helper.build_packet('GP', ID)
        self.communicator.write(command_line)
        time.sleep(0.1)
        data_buffer = self.read_untils_have_data('ID')

        if data_buffer is None:
            return False

        serial_num = int.from_bytes(struct.pack(
            '4B', *data_buffer[0:4]), byteorder='big')

        mode_string_len = len(data_buffer[4:])
        model_string = struct.pack('{0}B'.format(
            mode_string_len), *data_buffer[4:]).decode()

        split_text = model_string.split(' ')
        if model_string.find('OpenIMU') > -1 or \
                model_string.find('OpenRTK') > -1:
            return False

        self.device_info = {
            'name': split_text[0],
            'pn': split_text[1],
            # 'firmware_version': split_text[2],
            'sn': serial_num
        }
        return True

    def _build_app_info(self):
        '''
        Build app info
        '''
        command_line = dmu_helper.build_packet('GP', VR)
        self.communicator.write(command_line)
        time.sleep(0.1)
        data_buffer = self.read_untils_have_data('VR')

        if data_buffer is None:
            return False

        version_string = '{0}.{1}.{2}.{3}.{4}'.format(*data_buffer)

        self.app_info = {
            'app_name': 'DMU',
            'version': version_string
        }
        return True

    def load_properties(self):
        # Load config from user working path
        local_config_file_path = os.path.join(os.getcwd(), 'dmu.json')
        if os.path.isfile(local_config_file_path):
            with open(local_config_file_path) as json_data:
                self.properties = json.load(json_data)
                return

        app_file_path = os.path.join(
            self.setting_folder_path, 'dmu.json')

        with open(app_file_path) as json_data:
            self.properties = json.load(json_data)
            CONFIGURATION_FIELD_DEFINES_SINGLETON.load(
                self.properties['userConfiguration'])
            EEPROM_FIELD_DEFINES_SINGLETON.load()

    def after_setup(self):
        pass

    def on_read_raw(self, data):
        pass

    def on_receive_output_packet(self, packet_type, data):
        '''
        Listener for getting output packet
        '''
        self.add_output_packet('stream', packet_type, data)

    def get_device_info(self, *args):  # pylint: disable=unused-argument
        '''
        Get device information
        '''
        return {
            'packetType': 'deviceInfo',
            'data':  [
                {'name': 'Product Name',
                 'value': self.device_info['name']},
                {'name': 'PN', 'value': self.device_info['pn']},
                {'name': 'Firmware Version',
                 'value': self.app_info['version']},
                {'name': 'SN', 'value': self.device_info['sn']}
            ]
        }

    @with_device_message
    def get_conf(self, *args):  # pylint: disable=unused-argument
        '''
        Get json configuration
        '''
        outputs = self.properties['userMessages']['outputPackets']
        input_params = self.properties['userConfiguration']

        if self.is_conf_loaded:
            yield {
                'packetType': 'conf',
                'data': {
                    'outputs': outputs,
                    'inputParams': input_params
                }
            }

        # read product configuration
        eeprom_field = EEPROM_FIELD_DEFINES_SINGLETON.find(0x71C)
        command_line = dmu_helper.build_read_eeprom_cli(eeprom_field)
        result = yield self._message_center.build(command=command_line, timeout=3)

        data = result['data']

        if data:
            packet_types = dmu_helper.build_continous_packet_types(
                data['value']['architechture'],
                data['value']['algorithm'],
                data['value']['mags'])
            for item in input_params:
                if item['name'] == 'Packet Type':
                    # product_configuration['continuous_packet_types']
                    item['options'] = packet_types
                    self.is_conf_loaded = True
                    break

        yield {
            'packetType': 'conf',
            'data': {
                'outputs': outputs,
                'inputParams': input_params
            }
        }

    @with_device_message
    def get_params(self, *args):  # pylint: disable=unused-argument
        '''
        Get all parameters
        '''
        fields = CONFIGURATION_FIELD_DEFINES_SINGLETON.get_fields()
        command_line = dmu_helper.build_read_fields_packets(fields)
        result = yield self._message_center.build(command=command_line, timeout=3)

        data = result['data']

        if data:
            self.parameters = data
            yield {
                'packetType': 'inputParams',
                'data': data
            }
        yield {
            'packetType': 'error',
            'data': 'No Response'
        }

    @with_device_message
    def get_param(self, params, *args):  # pylint: disable=unused-argument
        '''
        Update paramter value
        '''
        field = CONFIGURATION_FIELD_DEFINES_SINGLETON.find(params['paramId'])

        command_line = dmu_helper.build_read_fields_packets([field])

        result = yield self._message_center.build(command=command_line)

        data = result['data']
        if data:
            yield {
                'packetType': 'inputParam',
                'data': data
            }
        yield {
            'packetType': 'error',
            'data': 'No Response'
        }

    @with_device_message
    def set_params(self, params, *args):  # pylint: disable=unused-argument
        '''
        Update paramters value
        '''
        for parameter in params:
            command_line = helper.build_input_packet(
                'uP', properties=self.properties,
                param=parameter['paramId'],
                value=parameter['value'])

            result = yield self._message_center.build(command=command_line)

            packet_type = result['packet_type']
            data = result['data']

            if packet_type == 'error':
                yield {
                    'packetType': 'error',
                    'data': {
                        'error': data
                    }
                }
            if data > 0:
                yield {
                    'packetType': 'error',
                    'data': {
                        'error': data
                    }
                }

        yield {
            'packetType': 'success',
            'data': {
                'error': 0
            }
        }

    @with_device_message
    def set_param(self, params, *args):  # pylint: disable=unused-argument
        '''
        Update paramter value
        '''
        configuration_field = CONFIGURATION_FIELD_DEFINES_SINGLETON.find(
            params['paramId'])
        command_line = dmu_helper.build_write_filed_cli(
            configuration_field, params['value'])

        result = yield self._message_center.build(command=command_line)

        error = result['error']
        data = result['data']
        if error:
            yield {
                'packetType': 'error',
                'data': {
                    'error': error
                }
            }

        yield {
            'packetType': 'success',
            'data': {
                'error': data
            }
        }

    @with_device_message
    def save_config(self, *args):  # pylint: disable=unused-argument
        '''
        Save configuration
        '''
        # read current configuration fields, then write field to eeprom
        fields = CONFIGURATION_FIELD_DEFINES_SINGLETON.get_fields()
        command_line = dmu_helper.build_read_fields_packets(fields)
        result = yield self._message_center.build(command=command_line, timeout=3)

        data = result['data']
        values = [item['value'] for item in data]
        print('saved values', values)

        command_line = dmu_helper.build_write_fileds_cli(fields, values, True)
        result = yield self._message_center.build(command=command_line, timeout=3)

        data = result['data']
        error = result['error']
        if data:
            yield {
                'packetType': 'success',
                'data': data
            }

        yield {
            'packetType': 'success',
            'data': error
        }



from typing import List
from ..parsers.dmu_field_parser import (encode_value, decode_value)


class ConfigurationField(object):
    '''
    Configuration Field
    '''

    def __init__(self, name, field_id, field_type):
        self.name = name
        self.field_id = field_id
        self.field_type = field_type

    def parse(self, payload):
        '''
        Parse payload
        '''
        # return: value, parsed, error
        value = decode_value(self.field_type, payload)
        return value, True, None

    def encode(self, value):
        return encode_value(self.field_type, value)


class ConfigruationFieldDefines:
    '''
    A list of fields could be parsed
    '''

    def __init__(self):
        self._list = dict()

    def load(self, json):
        self._list.clear()
        for item in json:
            param_id = item['paramId']
            param_name = item['name']
            param_type = item['type']
            self._list[param_id] = ConfigurationField(
                param_name, param_id, param_type)

    def get_fields(self) -> List[ConfigurationField]:
        return self._list.values()

    def find(self, paramId: int) -> ConfigurationField:
        return self._list.get(paramId)


CONFIGURATION_FIELD_DEFINES_SINGLETON = ConfigruationFieldDefines()



import struct


def decode_value(data_type, data):
    if data_type == 'uint64':
        try:
            pack_item = struct.pack('8B', *data)
        except:  # pylint: disable=bare-except
            return False
        return struct.unpack('<Q', pack_item)[0]
    elif data_type == 'int64':
        try:
            pack_item = struct.pack('8B', *data)
        except:  # pylint: disable=bare-except
            return False
        return struct.unpack('<q', pack_item)[0]
    elif data_type == 'double':
        try:
            pack_item = struct.pack('8B', *data)
        except:  # pylint: disable=bare-except
            return False
        return struct.unpack('d', pack_item)[0]
    elif data_type == 'uint32':
        try:
            pack_item = struct.pack('4B', *data)
        except:  # pylint: disable=bare-except
            return False
        return struct.unpack('<I', pack_item)[0]
    elif data_type == 'int32':
        try:
            pack_item = struct.pack('4B', *data)
        except:  # pylint: disable=bare-except
            return False
        return struct.unpack('<i', pack_item)[0]
    elif data_type == 'float':
        try:
            pack_item = struct.pack('4B', *data)
        except:  # pylint: disable=bare-except
            return False
        return struct.unpack('<f', pack_item)[0]
    elif data_type == 'uint16':
        try:
            pack_item = struct.pack('2B', *data)
        except:  # pylint: disable=bare-except
            return False
        return struct.unpack('<H', pack_item)[0]
    elif data_type == 'int16':
        try:
            pack_item = struct.pack('2B', *data)
        except:  # pylint: disable=bare-except
            return False
        return struct.unpack('<h', pack_item)[0]
    elif data_type == 'uint8':
        try:
            pack_item = struct.pack('1B', *data)
        except:  # pylint: disable=bare-except
            return False
        return struct.unpack('<B', pack_item)[0]
    elif data_type == 'int8':
        try:
            pack_item = struct.pack('1B', *data)
        except:  # pylint: disable=bare-except
            return False
        return struct.unpack('<b', pack_item)[0]
    elif 'char' in data_type:
        try:
            trim_data = [elem for elem in data if elem != 0]
            max_len = int(data_type.replace('char', ''))
            trim_len = len(trim_data)
            data_len = trim_len if trim_len < max_len else max_len
            pack_item = struct.pack('{0}B'.format(data_len), *trim_data)
            return pack_item.decode()
        except:  # pylint: disable=bare-except
            return False
    elif data_type == 'string':
        try:
            fmt_str = '{0}B'.format(len(data))
            pack_item = struct.pack(fmt_str, *data)
            return pack_item
        except:  # pylint: disable=bare-except
            return False
    elif data_type == 'ip4':
        try:
            ip_1 = str(data[0])
            ip_2 = str(data[1])
            ip_3 = str(data[2])
            ip_4 = str(data[3])
            return ip_1+'.'+ip_2+'.'+ip_3+'.'+ip_4
        except:  # pylint: disable=bare-except
            return False
    elif data_type == 'ip6':
        try:
            ip_1 = str(data[0])
            ip_2 = str(data[1])
            ip_3 = str(data[2])
            ip_4 = str(data[3])
            ip_5 = str(data[4])
            ip_6 = str(data[5])
            return ip_1+'.'+ip_2+'.'+ip_3+'.'+ip_4+'.'+ip_5+'.'+ip_6
        except:  # pylint: disable=bare-except
            return False
    else:
        return False



import collections
import operator
from ..base.event_base import EventBase
from ...framework.utils import helper
from .dum_packet_parser import (
    match_command_handler, match_continuous_handler)

MSG_HEADER = [0x55, 0x55]
NAK = [0x15, 0x15]
PACKET_TYPE_INDEX = 2
PRIVATE_PACKET_TYPE = ['RE', 'WE', 'UE', 'LE', 'SR']
INPUT_PACKETS = ['PK', 'CH', 'RE', 'WE', 'UE', 'GP',
                 'SF', 'RF', 'WF', 'GF',
                 'AR', 'SR', 'PR', '\x15\x15', '\x00\x00',
                 'WC', 'CB', 'CC']
OUTPUT_PACKETS = ['ID', 'VR', 'VA', 'KC', 'KT', 'KS']


class UartMessageParser(EventBase):
    def __init__(self, configuration):
        super(UartMessageParser, self).__init__()
        self.frame = []
        self.payload_len_idx = 5
        self.sync_pattern = collections.deque(2*[0], 2)
        self.find_header = False
        self.payload_len = 0
        self.properties = configuration
        self.run_command = ''
        # command,continuous_message

    def set_configuration(self, configuration):
        self.properties = configuration

    def set_run_command(self, command):
        self.run_command = ''.join(['%c' % x for x in command[2:4]])

    def analyse(self, data_block):
        if self.find_header:
            self.frame.append(data_block)
            if self.payload_len_idx == len(self.frame):
                self.payload_len = data_block

            elif 5 + self.payload_len + 2 == len(self.frame):
                packet_type = ''.join(
                    ["%c" % x for x in self.frame[PACKET_TYPE_INDEX:4]])
                self.find_header = False
                result = helper.calc_crc(self.frame[2:-2])
                if result[0] == self.frame[-2] and result[1] == self.frame[-1]:
                    # find a whole frame
                    # self._parse_frame(self.frame, self.payload_len)
                    self._parse_message(
                        packet_type, self.payload_len, self.frame[5:self.payload_len+5])

                    self.find_header = False
                    self.payload_len = 0
                    self.sync_pattern = collections.deque(2*[0], 2)
                else:
                    print("crc check error! packet_type:", packet_type)
                    input_packet_config = next(
                        (x for x in self.properties['userMessages']['inputPackets']
                         if x['name'] == packet_type), None)
                    if input_packet_config:
                        self.emit('command', packet_type=packet_type,
                                  data=[], error=True)

        else:
            self.sync_pattern.append(data_block)
            if operator.eq(list(self.sync_pattern), MSG_HEADER):
                self.frame = MSG_HEADER[:]  # header_tp.copy()
                self.find_header = True

    def _parse_message(self, packet_type, payload_len, payload):
        # parse interactive commands
        is_interactive_cmd = INPUT_PACKETS.__contains__(packet_type)
        if is_interactive_cmd:
            self._parse_input_packet(packet_type, payload_len, payload)
        else:
            # consider as output packet, parse output Messages
            self._parse_output_packet(packet_type, payload_len, payload)

    def _parse_input_packet(self, packet_type, payload_len, payload):
        # print(packet_type, payload, payload_len)
        payload_parser = match_command_handler(packet_type)

        if payload_parser:
            data, error = payload_parser(payload)

            self.emit('command',
                      packet_type=packet_type,
                      data=data,
                      error=error)
        else:
            print('[Warning] Unsupported command {0}'.format(packet_type))

    def _parse_output_packet(self, packet_type, payload_len, payload):
        # check if it is the valid out packet
        payload_parser = match_continuous_handler(packet_type)
        output_packet_config = next(
            (x for x in self.properties['userMessages']['outputPackets']
             if x['name'] == packet_type), None)
        scaling = self.properties['scaling']

        data = payload_parser(payload, output_packet_config, scaling)

        if not data:
            print('Cannot parse packet type {0}'.format(packet_type))
            return

        if self.run_command == 'GP':
            self.emit('command',
                      packet_type=packet_type,
                      data=data)
            self.run_command = ''
        else:
            self.emit('continuous_message',
                      packet_type=packet_type,
                      data=data)



class KeyValuesArgumentBase(object):
    '''
    Object Argument Base
    '''
    default_values = {}

    def __init__(self, **kwargs):
        keys = self.default_values.keys()
        for key in keys:
            value = self._prepare_value(kwargs, key)
            setattr(self, key, value)

    def _prepare_value(self, input_args, key):
        value = input_args.get(key)
        if not value:
            value = self.default_values.get(key)
        return value

    def __str__(self):
        str_out = ''
        keys = self.default_values.keys()
        for key in keys:
            value = getattr(self, key)
            str_out = str_out+str(key)+"="+str(value)+" "
        return str_out


class WebserverArgs(KeyValuesArgumentBase):
    '''
    Argument define for start webserver
    '''
    default_values = {
        'device_type': 'auto',
        'port': 'auto',
        'baudrate': 'auto',
        'com_port': 'auto',
        'debug': False,
        'with_data_log': False,
        'with_raw_log': False,
        'console_log': False,
        'set_user_para': False,
        'ntrip_client': False
    }


class DetectorArgs(KeyValuesArgumentBase):
    '''
    Argument define for detect device
    '''
    default_values = {
        'device_type': 'auto',
        'baudrate': 'auto',
        'com_port': 'auto'
    }



"""
Communicator
"""
import sys
import os
import time
import json
import serial
import serial.tools.list_ports
from ..devices import DeviceManager
from .constants import BAUDRATE_LIST
from .context import APP_CONTEXT
from .utils.resource import (
    get_executor_path
)
if sys.version_info[0] > 2:
    from queue import Queue
else:
    from Queue import Queue


class CommunicatorFactory:
    '''
    Communicator Factory
    '''
    @staticmethod
    def create(method, options):
        '''
        Initial communicator instance
        '''
        if method == 'uart':
            return SerialPort(options)
        elif method == 'spi':
            return SPI(options)
        else:
            raise Exception('no matched communicator')


class Communicator(object):
    '''Communicator base
    '''

    def __init__(self):
        executor_path = get_executor_path()
        setting_folder_name = 'setting'
        self.setting_folder_path = os.path.join(
            executor_path, setting_folder_name)
        self.connection_file_path = os.path.join(
            self.setting_folder_path, 'connection.json')
        self.read_size = 0
        self.device = None

    def find_device(self, callback):
        '''
        find device, then invoke callback
        '''
        callback()

    def open(self):
        '''
        open
        '''

    def close(self):
        '''
        close
        '''

    def write(self, data, is_flush=False):
        '''
        write
        '''

    def read(self, size):
        '''
        read
        '''

    def confirm_device(self, *args):
        '''
        validate the connected device
        '''
        self.device = DeviceManager.ping(self, *args)


class SerialPort(Communicator):
    '''
    Serial Port
    '''

    def __init__(self, options=None):
        super(SerialPort, self).__init__()
        self.type = 'uart'
        self.serial_port = None  # the active UART
        self.port = None
        self.baud = None
        self.read_size = 100
        self.baudrate_assigned = False
        # self.baudrateList = [115200]  # for test
        self.baudrate_list = BAUDRATE_LIST  # default baudrate list
        self.com_port = None
        self.com_port_assigned = False
        self.filter_device_type = None
        self.filter_device_type_assigned = False

        if options and options.baudrate != 'auto':
            self.baudrate_list = [options.baudrate]
            self.baudrate_assigned = True
        if options and options.com_port != 'auto':
            self.com_port = options.com_port
            self.com_port_assigned = True
        if options and options.device_type != 'auto':
            self.filter_device_type = options.device_type
            self.filter_device_type_assigned = True

    def find_device(self, callback):
        ''' Finds active ports and then autobauds units
        '''
        self.device = None
        if self.com_port_assigned:
            # find device by assigned port
            self.autobaud([self.com_port])
            if self.device is None:
                raise Exception(
                    'Cannot open serial port: {0}'.format(self.com_port))
        else:
            while self.device is None:
                if self.try_last_port():
                    break

                num_ports = self.find_ports()
                self.autobaud(num_ports)
                time.sleep(0.5)
        callback(self.device)

    def find_ports(self):
        '''
        Find available ports
        '''
        port_list = list(serial.tools.list_ports.comports())
        ports = [p.device for p in port_list]

        result = []

        for port in ports:
            if "Bluetooth" in port:
                continue
            else:
                # print('Check if is a used port ' + port)
                ser = None
                try:
                    ser = serial.Serial(port, exclusive=True)
                    if ser:
                        ser.close()
                        result.append(port)
                except Exception as ex:
                    APP_CONTEXT.get_logger().logger.debug(
                        'actual port exception %s', ex)
                    APP_CONTEXT.get_logger().logger.info(
                        'port:%s is in use', port)
        return result

    def autobaud(self, ports):
        '''Autobauds unit - first check for stream_mode/continuous data, then check by polling unit
           Converts resets polled unit (temporarily) to 100Hz ODR
           :returns:
                true when successful
        '''
        APP_CONTEXT.get_logger().logger.info('start to connect serial port')
        baudrate_list_from_options = self.baudrate_list

        for port in ports:
            for baud in baudrate_list_from_options:
                APP_CONTEXT.get_logger().logger.info(
                    "try {0}:{1}".format(port, baud))
                self.open(port, baud)
                if self.serial_port is not None:
                    self.confirm_device(self.filter_device_type)

                    if self.device is None:
                        self.close()
                        continue
                    else:
                        self.save_last_port()
                        return True
        return False

    def try_last_port(self):
        '''try to open serial port based on the port and baud read from connection.json.
           try to find frame header in serial data.
           returns: True if find header
                    False if not find header.
        '''
        connection = None
        try:
            if not os.path.isfile(self.connection_file_path):
                return False

            with open(self.connection_file_path) as json_data:
                connection = json.load(json_data)
            connection['baud'] = self.baudrate_list[0] if self.baudrate_assigned \
                else connection['baud']
            APP_CONTEXT.get_logger().logger.info('try to use last connected port {} {}'.format(
                connection['port'], connection['baud']))
            if connection:
                self.open_serial_port(
                    port=connection['port'], baud=connection['baud'], timeout=0.005)
                if self.serial_port is not None:
                    self.confirm_device(self.filter_device_type)
                    if self.device is None:
                        self.close()
                        return False
                    else:
                        self.save_last_port()
                        # Assume max_len of a frame is less than 300 bytes.
                        return True
                else:
                    return False
        except Exception as ex:
            print(ex)
            return False

    def save_last_port(self):
        '''
        save connected port info
        '''

        if not os.path.exists(self.setting_folder_path):
            try:
                os.mkdir(self.setting_folder_path)
            except:
                return

        connection = {"port": self.serial_port.port,
                      "baud": self.serial_port.baudrate}
        try:
            with open(self.connection_file_path, 'w') as outfile:
                json.dump(connection, outfile)
        except:
            pass

    def open_serial_port(self, port=None, baud=115200, timeout=0.1):
        ''' open serial port
            returns: true when successful
        '''
        try:
            self.serial_port = serial.Serial(
                port, baud, timeout=timeout, exclusive=True)
            return True
        except Exception as ex:
            APP_CONTEXT.get_logger().logger.info(
                '{0} : {1} open failed'.format(port, baud))
            if self.serial_port is not None:
                if self.serial_port.isOpen():
                    self.serial_port.close()

            self.serial_port = None
            return False

    def close_serial_port(self):
        '''close serial port
        '''
        if self.serial_port is not None:
            if self.serial_port.isOpen():
                self.serial_port.close()

    def write(self, data, is_flush=False):
        '''
        write the bytes data to the port

        return:
                length of data sent via serial port.
                False: Exception when sending data, eg. serial port hasn't been opened.
        '''
        try:
            len_of_data = self.serial_port.write(data)
            if is_flush:
                self.serial_port.flush()
            return len_of_data
        except Exception as ex:
            # print(e)
            raise

    def read(self, size=100):
        '''
        read size bytes from the serial port.
        parameters: size - number of bytes to read.
        returns: bytes read from the port.
        return type: bytes
        '''
        try:
            return self.serial_port.read(size)
        except serial.SerialException:
            print(
                'Serial Exception! Please check the serial port connector is stable or not.')
            raise
        except Exception as ex:
            # print(e)
            raise

    def open(self, port=False, baud=57600):
        return self.open_serial_port(port, baud, timeout=0.005)

    def close(self):
        return self.close_serial_port()

    def reset_buffer(self):
        '''
        reset buffer
        '''
        self.serial_port.flushInput()
        self.serial_port.flushOutput()


class SPI(Communicator):
    '''SPI'''

    def __init__(self, options=None):
        super().__init__()
        self.type = 'spi'



# Copyright (C) 2018 Aceinna Navigation System Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.

from setuptools import find_packages, setup
from src.aceinna import (PACKAGE_NAME, VERSION)

PACKAGE_DESCRIPTION = "Aceinna Navigation System Open Devices Library"

INSTALL_REQUIRES = [
    "pyserial",
    "pathlib",
    "psutil",
    "azure-storage-blob==2.1.0",
    "tornado"
]

setup(
    name=PACKAGE_NAME,
    version=VERSION,
    author="Aceinna, Inc",
    author_email="info@aceinna.com",
    description=PACKAGE_DESCRIPTION,
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Aceinna/python-openimu",
    license="Apache 2.0",
    python_requires=">=2.7, !=3.0.*, !=3.1.*",
    install_requires=INSTALL_REQUIRES,
    packages=find_packages("src", exclude=['test', 'tests']),
    package_dir={"": "src"},
    package_data={
        'aceinna': ['setting/openrtk/*/*.json', 'setting/openimu/*/*.json', 'setting/dmu/*.json']
    },
    classifiers=[
        "Environment :: Console",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    entry_points={
        "console_scripts": [
            "openimu = aceinna.tools.cli:main",
        ]
    }
)
